####################################################################################################
# neuropythy/util/labels.py
# Simple tools for dealing with neuroscience-related labels for brains.
# By Noah C. Benson

import numpy               as np
import pyrsistent          as pyr
import collections         as colls
import os, sys, types, six, pimms

from .core import (ObjectWithMetaData, is_tuple, curry, is_dataframe)

def label_colors(lbls, cmap=None):
    '''
    label_colors(labels) yields a dict object whose keys are the unique values in labels and whose
      values are the (r,g,b,a) colors that should be assigned to each label.
    label_colors(n) is equivalent to label_colors(range(n)).

    Note that this function uses a heuristic and is not guaranteed to be optimal in any way for any
    value of n--but it generally works well enough for most common purposes.
    
    The following optional arguments may be given:
      * cmap (default: None) specifies a colormap to use as a base. If this is None, then a varianct
        of 'hsv' is used.
    '''
    from neuropythy.graphics import label_cmap
    if pimms.is_int(lbls): lbls = np.arange(lbls)
    lbls0 = np.unique(lbls)
    lbls = np.arange(len(lbls0))
    cm = label_cmap(lbls, cmap=cmap)
    mx = float(len(lbls) - 1)
    m = {k:cm(l/mx) for (k,l) in zip(lbls0, lbls)}
    return m
@pimms.immutable
class LabelEntry(ObjectWithMetaData):
    '''
    LabelEntry is a class tracked by LabelIndex objects; it stores information about a single
    label.
    '''
    def __init__(self, ident, name, color=None, meta_data=None):
        self.id = ident
        self.name = name
        self.color = color
        self.meta_data = meta_data
    @pimms.param
    def id(i):
        'le.id is the id of the given label entry object le.'
        if not pimms.is_int(i): raise ValueError('label-entry id must be an int')
        return int(i)
    @pimms.param
    def name(nm):
        'le.name is the (string) name of the given label entry object le.'
        if not pimms.is_str(nm): raise ValueError('label-entry name must be a string')
        return nm
    @pimms.param
    def color(c):
        'le.color is the tuple (r,g,b,a) for the given label entry object le.'
        if c is None: return c
        c = tuple(c)
        if len(c) == 3: c = c + (1,)
        if len(c) != 4: raise ValueError('Invalid color: %s' % c)
        return c
    def __repr__(self):
        return 'label(<%d: %s>)' % (self.id, self.name)
@pimms.immutable
class LabelIndex(ObjectWithMetaData):
    '''
    LabelIndex is an immutable class that tracks label data and can lookup labels by name or integer
    value as well as assign colors to them.
    '''
    def __init__(self, ids, names, colors=None, entry_meta_data=None, meta_data=None):
        '''
        LabelIndex(ids, names) constructs a label index object with the given ids and names.
        '''
        self.ids = ids
        self.names = names
        self.colors = colors
        self.entry_meta_data = None
        self.meta_data = meta_data
    @pimms.param
    def ids(ii):
        '''
        lblidx.ids is a tuple of the integer identifiers used for the labels in the given label
        index object.
        '''
        return tuple(ii)
    @pimms.param
    def names(nms):
        '''
        lblidx.names is a tuple of names used for the labels in the given label index object.
        '''
        return tuple(nms)
    @pimms.param
    def colors(cs):
        '''
        lblidx.colors is a numpy array of colors for each label or None.
        '''
        from neuropythy.graphics import to_rgba
        if cs is None: return None
        # we want to convert to colors
        return pimms.imm_array([to_rgba(c) for c in cs])
    @pimms.param
    def entry_meta_data(mds):
        '''
        lblidx.entry_meta_data is lists of meta-data maps for each of the labels in the given label
          index object.
        '''
        if mds is None: return None
        if is_dataframe(mds):
            mds = {k:mds[k].values for k in mds.colums}
        elif pimms.is_map(mds):
            ks = list(mds.keys())
            mds = [{k:v for (k,v) in zip(ks,vs)} for vs in np.transpose(list(mds.values()))]
        elif not pimms.is_array(mds) or not all(pimms.is_map(u) for u in mds):
            raise ValueError('unbalanced or non-map entry meta-data')
        return pimms.persist(mds)
    @pimms.require
    def check_counts(ids, names, colors, entry_meta_data):
        '''
        Checks that ids, names, and colors are the same length and that ids is unique.
        '''
        if len(np.unique(ids)) != len(ids): raise ValueError('label index ids must be unique')
        if len(ids) != len(names): raise ValueError('label index names and ids must be same length')
        if colors is not None and len(colors) != len(ids):
            raise ValueError('label index colors and ids must be same length')
        if entry_meta_data is not None and len(entry_meta_data) != len(ids):
            raise ValueError('label index entry_meta_data and ids must be same length')
        return True
    @pimms.value
    def entries(ids, names, colors, entry_meta_data):
        '''
        lblidx.entries is a tuple of the label entry objects for the given label index object.
        '''
        if 0 not in ids:
            ids = np.concatenate([[0],ids])
            names = ['none'] + list(names)
            if colors          is not None: colors = np.vstack([[(0,0,0,0)], colors])
            if entry_meta_data is not None: entry_meta_data = [None] + list(entry_meta_data)
        if colors is None: colors = np.asarray([cs[k] for cs in [label_colors(ids)] for k in ids])
        if entry_meta_data is None: entry_meta_data = [None]*len(ids)
        # the 0 id is implied if not given:
        les = [LabelEntry(ii, name, color=color, meta_data=md).persist()
               for (ii,name,color,md) in zip(ids, names, colors, entry_meta_data)]
        return tuple(les)
    @pimms.value
    def by_id(entries):
        '''
        lblidx.by_id is a persistent map of the label entries indexed by their identifier.
        '''
        return pyr.pmap({e.id:e for e in entries})
    @pimms.value
    def by_name(entries):
        '''
        lblidx.by_name is a persistent map of the label entries indexed by their names.
        '''
        return pyr.pmap({e.name:e for e in entries})
    @pimms.value
    def vmin(ids):
        '''
        lblidx.vmin is the minimum value of a label identifier in the given label index.
        '''
        return np.min(ids)
    @pimms.value
    def vmax(ids):
        '''
        lblidx.vmax is the maximum value of a label identifier in the given label index.
        '''
        return np.max(ids)
    @pimms.value
    def colormap(entries):
        '''
        lblidx.colormap is a colormap appropriate for use with data that has been scaled to run from
        0 at lblidx.vmin to 1 at lblidx.vmax.
        '''
        import matplotlib.colors
        from_list = matplotlib.colors.LinearSegmentedColormap.from_list
        ids = np.asarray([e.id for e in entries])
        ii  = np.argsort(ids)
        ids = ids[ii]
        clrs = np.asarray([e.color for e in entries])[ii]
        (vmin,vmax) = [f(ids) for f in (np.min, np.max)]
        vals = (ids - vmin) / (vmax - vmin)
        return from_list('label%d' % len(vals), list(zip(vals, clrs)))
    def __getitem__(self, k):
        if   pimms.is_int(k):              return self.by_id.get(k, None)
        elif pimms.is_str(k):              return self.by_name.get(k, None)
        elif pimms.is_vector(k, 'int'):    return np.asarray([self.by_id.get(k, None) for k in k])
        else: return np.asarray([(self.by_name if pimms.is_str(k) else self.by_id).get(k, None)
                                 for k in k])
    def name_lookup(self, ii):
        '''
        lblidx.name_lookup(ii) yields the names associated with the labels with the given ids. If
          ii is a list of ids, then yields an array of names.
        '''
        if   pimms.is_int(ii): return self.by_id[ii].name if ii in self.by_id else None
        elif pimms.is_str(ii): return self.by_name[ii].name if ii in self.by_name else None
        else: return np.asarray([tbl[ii].name if ii in tbl else None
                                 for ii in ii
                                 for tbl in [self.by_name if pimms.is_str(ii) else self.by_id]])
    def id_lookup(self, names):
        '''
        lblidx.id_lookup(names) yields the ids associated with the labels with the given names. If
          names is a list of names, then yields an array of ids.
        '''
        if   pimms.is_str(names): return self.by_name[names].id if names in self.by_name else None
        elif pimms.is_int(names): return self.by_id[names].id if names in self.by_id else None
        else: return np.asarray([tbl[ii].id if ii in tbl else None
                                 for ii in names
                                 for tbl in [self.by_name if pimms.is_str(ii) else self.by_id]])
    def color_lookup(self, ii):
        '''
        lblidx.color_lookup(ids) yields the color(s) associated with the labels with the given ids.
          If ids is a list of ids, then yields a matrix of colors.
        lblidx.color_lookup(names) uses the names to lookup the label colors.
        '''
        if pimms.is_int(ii): return self.by_id[ii].color if ii in self.by_id else None
        elif pimms.is_str(ii): return self.by_name[ii].color if ii in self.by_name else None
        else: return np.asarray([tbl[ii].color if ii in tbl else None
                                 for ii in ii
                                 for tbl in [self.by_name if pimms.is_str(ii) else self.by_id]])
    def cmap(self, data=None):
        '''
        lblidx.cmap() yields a colormap for the given label index object that assumes that the data
          being plotted will be rescaled such that label 0 is 0 and the highest label value in the
          label index is equal to 1.
        lblidx.cmap(data) yields a colormap that will correctly color the labels given in data if
          data is scaled such that its minimum and maximum value are 0 and 1.
        '''
        import matplotlib.colors
        from_list = matplotlib.colors.LinearSegmentedColormap.from_list
        if data is None: return self.colormap
        data = np.asarray(data).flatten()
        (vmin,vmax) = (np.min(data), np.max(data))
        ii  = np.argsort(self.ids)
        ids = np.asarray(self.ids)[ii]
        if vmin == vmax:
            (vmin,vmax,ii) = (vmin-0.5, vmax+0.5, vmin)
            clr = self.color_lookup(ii)
            return from_list('label1', [(0, clr), (1, clr)])
        q   = (ids >= vmin) & (ids <= vmax)
        ids = ids[q]
        clrs = self.color_lookup(ids)
        vals = (ids - vmin) / (vmax - vmin)
        return from_list('label%d' % len(vals), list(zip(vals, clrs)))
    def __repr__(self):
        return 'LabelIndex(<%d labels>)' % len(self.ids)
def is_label_index(le):
    '''
    is_label_index(le) yields True if the given object le is a label index and False otherwise.
    '''
    return isinstance(le, LabelIndex)
def label_index(dat, *args, **kw):
    '''
    label_index(idx_map) converts the given map- or dict-like object idx_map into a label index by
      assuming that the keys are label ids and the values are label names or tuples of label names
      and (r,g,b,a) colors.
    label_index(ids, names) uses the given ids and names to make the label index.
    label_index(ids, names, colors) additionally uses the given colors.

    Note that if there is not a label with id 0 then such a label is automatically created with the
    name 'none', the rgba color [0,0,0,0], and no entry meta-data. As a general rule, the label 0
    should be used to indicate that a label is missing.

    The optional arguments meta_data and entry_meta_data may specify both the meta-data for the
    label-index object itself as well as the meta-data for the individual entries.
    '''
    md = kw.pop('meta_data', {})
    mds = kw.pop('entry_meta_data', None)
    if len(kw) > 0: raise ValueError('unrecognized optional argument(s) given to label_index')
    if len(args) == 0:
        if pimms.is_map(dat):
            (ids,nms,clrs) = ([],[],[])
            for (k,v) in six.iteritems(dat):
                if pimms.is_scalar(v): c = None
                else: (v,c) = v
                if pimms.is_str(k):
                    ids.append(v)
                    nms.append(k)
                else:
                    ids.append(k)
                    nms.append(v)
                if c is not None: clrs.append(c)
        elif is_dataframe(dat):
            if dat.index.name.lower() == 'id': ids = dat.index.values
            else: ids = dat['id'].values
            nms = dat['name'].values
            if 'color' in dat: clrs = np.array(list(map(list, dat['color'].values)))
            elif all(k in dat for k in ['r','g','b']):
                ks = ['r','g','b']
                if 'a' in dat: ks.append('a')
                clrs = np.array([[r[k] for k in ks].values for (ii,r) in dat.iterrows()])
            else: clrs = []
        elif pimms.is_vector(dat, 'int'):
            ids = np.unique(dat)
            nms = ['label%d'%k for k in ids]
            clrs = []
        else: raise ValueError('label_index(idx_map) given non-map argument')
    elif len(args) == 1: (ids,nms,clrs) = (dat, args[0], [])
    elif len(args) == 2: (ids,nms,clrs) = (dat, args[0], args[1])
    else: raise ValueError('Too many arguments given to label_index()')
    if clrs is None or len(clrs) == 0: clrs = None
    elif len(clrs) != len(ids): raise ValueError('color-count must match id-count')
    # okay, make the label index
    return LabelIndex(ids, nms, colors=clrs, meta_data=md, entry_meta_data=mds)
def to_label_index(obj):
    '''
    to_label_index(obj) attempts to coerce the given object into a label index object; if obj is
      already a label index object, then obj itself is returned. If obj cannot be coerced into a
      label index, then an error is raised.

    The obj argument can be any of the following:
      * a label index
      * a label list (i.e., an integer vector)
      * a tuple of arguments, potentially ending with a kw-options map, that can be passed to the
        label_index function successfully.
    '''
    if   is_label_index(obj): return obj
    elif pimms.is_vector(obj, 'int'): return label_index(obj)
    elif is_dataframe(obj): return label_index(obj)
    elif is_tuple(obj):
        if len(obj) > 1 and pimms.is_map(obj[-1]): return label_index(*obj[:-1], **obj[-1])
        else: return label_index(*obj)
    else: raise ValueError('could not parse to_label_index parameter: %s' % obj)
label_indices = {}



####################################################################################################
# main/benson2014_retinotopy.py
# The code for the function that handles the registration of retinotopy
# By Noah C. Benson

from __future__ import print_function

import numpy                        as     np
import scipy                        as     sp
import nibabel                      as     nib
import nibabel.freesurfer.io        as     fsio
import nibabel.freesurfer.mghformat as     fsmgh
import os, sys, six, pimms

from   ..freesurfer                 import (subject, add_subject_path)
from   ..vision                     import (predict_retinotopy, retinotopy_model, clean_retinotopy)
from   ..                           import io as nyio
from   ..mri                        import (is_image, to_image, image_clear)

info = \
   '''
   The benson14_retinotopy command can be used to project the anatomically defined
   template of retinotopy to a subject's left and right hemisphere(s).  At least
   one subject id (either a freesurfer subject name, if SUBJECTS_DIR is set
   appropriately in the environment, or a path to a subject directory) must be
   given. In each subject's freesurfer directory, a variety of output data is
   deposited:
    * surf/lh.benson14_angle  surf/rh.benson14_angle
      surf/lh.benson14_eccen  surf/rh.benson14_eccen
      surf/lh.benson14_varea  surf/rh.benson14_varea
      surf/lh.benson14_sigma  surf/rh.benson14_sigma
      These files contain predictions of polar angle, eccentricity, visual-area
      label, and pRF radius for each surface vertex in each hemisphere of the
      subject's hemispheres. The files are, by default, in FreeSurfer's curv
      format, but their format can be modified with the --surf-format flag.
    * mri/benson14_angle.mgz
      mri/benson14_eccen.mgz
      mri/benson14_varea.mgz
      mri/benson14_sigma.mgz
      These contain the data from the above surface data projected into the
      subject's 3D volume. Note that the volumes are oriented like Freesurfer's
      mri/brain.mgz file; if you want to convert this to the orientation of your
      original anatomical scan, use mri_convert:
       > mri_convert -rl mri/rawavg.mgz mri/angle_benson14.mgz \\
                     mri/scanner.angle_benson14.mgz
   The following options are accepted:
    * --eccen-tag=|-y<tag>
      --angle-tag=|-t<tag>
      --label-tag=|-l<tag>
      --sigma-tag=|-s<tag>
      These options specify the output tag to use for the predicted measurement
      that results from the registration. By default, these are
      'eccen_benson14', 'angle_benson14', 'varea_benson14', and 'sigma_benson14'.
      The output files have the name <hemi>.<tag>.mgz
    * --surf-format=|-o<nifti|nii.gz|nii|mgh|mgz|curv>
      --vol-format=|-v<nifti|nii.gz|nii|mgh|mgz>
      These flags specify what format the output should be in; note that nii.gz
      and nifti are identical; curv is a FreeSurfer curv (AKA morph data) file.
    * --no-volume-export|-x
      --no-surface-export|-z
      These flags indicate that the various data produced and written to the
      filesystem under normal execution should be suppressed. The volume export
      refers to the predicted volume files exported to the subject's mri directory
      and the surface export refers to the <hemi>.eccen_benson14.mgz and similar
      files that are written to the subject's surf directory.
    * --subjects-dir=|-d
      Specifies additional subject directory search locations (in addition to the
      SUBJECTS_DIR environment variable and the FREESURFER_HOME/subjects
      directories, which are given here in descending search priority) when looking
      for subjects by name. This option cannot be specified multiple times, but it
      may contain : characters to separate directories, as in PATH.
    * --no-overwrite|-n
      This flag indicates that, when writing output files, no file should ever be
      replaced, should it already exist.
    * --template=|-t
      Specifies the specific template that should be applied. By default this is
      'Benson14', the 2017 version of the template originally described in the paper
      by Benson et al. (2014). The option 'benson14' is also accepted. If the 
    * --reg=|-R<fsaverage|fsaverage_sym>
      Specifies the registration to look for the template in. This is, by default,
      fsaverage, but for the templates aligned to the fsaverage_sym hemisphere,
      this should specify fsaverage_sym.
    * --
      This token, by itself, indicates that the arguments that remain should not be
      processed as flags or options, even if they begin with a -.
   '''
_benson14_parser_instructions = [
    # Flags
    ('h', 'help',                   'help',              False),
    ('v', 'verbose',                'verbose',           False),
    ('x', 'no-volume-export',       'no_vol_export',     False),
    ('z', 'no-surface-export',      'no_surf_export',    False),
    ('n', 'no-overwrite',           'no_overwrite',      False),
    # Options
    ('e', 'eccen-tag',              'eccen_tag',         'benson14_eccen'),
    ('a', 'angle-tag',              'angle_tag',         'benson14_angle'),
    ('l', 'label-tag',              'label_tag',         'benson14_varea'),
    ('s', 'sigma-tag',              'sigma_tag',         'benson14_sigma'),
    ('d', 'subjects-dir',           'subjects_dir',      None),
    ('t', 'template',               'template',          'benson14'),
    ('o', 'surf-format',            'surf_format',       'curv'),
    ('v', 'vol-format',             'vol_format',        'mgz'),
    ('R', 'reg',                    'registration',      'fsaverage')]
_benson14_parser = pimms.argv_parser(_benson14_parser_instructions)

def main(*args):
    '''
    benson14_retinotopy.main(args...) runs the benson14_retinotopy command; see 
    benson14_retinotopy.info for more information.
    '''
    # Parse the arguments...
    (args, opts) = _benson14_parser(*args)
    # help?
    if opts['help']:
        print(info, file=sys.stdout)
        return 1
    # verbose?
    if opts['verbose']:
        def note(s):
            print(s, file=sys.stdout)
            return True
    else:
        def note(s): return False
    # based on format, how do we export?
    sfmt = opts['surf_format'].lower()
    if sfmt in ['curv', 'auto', 'automatic', 'morph']:
        sfmt = 'freesurfer_morph'
        sext = ''
    elif sfmt == 'nifti':
        sext = '.nii.gz'
    elif sfmt in ['mgh', 'mgz', 'nii', 'nii.gz']:
        sext = '.' + sfmt
    else:
        raise ValueError('Unknown surface format: %s' % opts['surf_format'])
    vfmt = opts['vol_format'].lower()
    if vfmt == 'nifti':
        vext = '.nii.gz'
    elif vfmt in ['mgh', 'mgz', 'nii', 'nii.gz']:
        vext = '.' + vfmt
    else:
        raise ValueError('Unknown volume format: %s' % opts['vol_format'])
    # Add the subjects directory, if there is one
    if 'subjects_dir' in opts and opts['subjects_dir'] is not None:
        add_subject_path(opts['subjects_dir'])
    ow = not opts['no_overwrite']
    nse = opts['no_surf_export']
    nve = opts['no_vol_export']
    tr = {'angle': opts['angle_tag'],
          'eccen': opts['eccen_tag'],
          'varea': opts['label_tag'],
          'sigma': opts['sigma_tag']}
    # okay, now go through the subjects...
    for subnm in args:
        note('Processing subject %s:' % subnm)
        sub = subject(subnm)
        note('   - Interpolating template...')
        (lhdat, rhdat) = predict_retinotopy(sub,
                                            template=opts['template'],
                                            registration=opts['registration'])
        # Export surfaces
        if nse:
            note('   - Skipping surface export.')
        else:
            note('   - Exporting surfaces:')
            for (t,dat) in six.iteritems(lhdat):
                flnm = os.path.join(sub.path, 'surf', 'lh.' + tr[t] + sext)
                if ow or not os.path.exists(flnm):
                    note('    - Exporting LH prediction file: %s' % flnm)
                    nyio.save(flnm, dat, format=sfmt)
                else:
                    note('    - Not overwriting existing file: %s' % flnm)
            for (t,dat) in six.iteritems(rhdat):
                flnm = os.path.join(sub.path, 'surf', 'rh.' + tr[t] + sext)
                if ow or not os.path.exists(flnm):
                    note('    - Exporting RH prediction file: %s' % flnm)
                    nyio.save(flnm, dat, format=sfmt)
                else:
                    note('    - Not overwriting existing file: %s' % flnm)
        # Export volumes
        if nve:
            note('   - Skipping volume export.')
        else:
            note('   - Exporting Volumes:')
            # generate the volumes once...
            im = sub.images['brain']
            note('    - Addressing volume...')
            addr = (sub.lh.image_address(im), sub.rh.image_address(im))
            for t in lhdat.keys():
                flnm = os.path.join(sub.path, 'mri', tr[t] + vext)
                if ow or not os.path.exists(flnm):
                    note('    - Preparing volume file: %s' % flnm)
                    dtyp = (np.int32 if t == 'varea' else np.float32)
                    vol = sub.cortex_to_image(
                        (lhdat[t], rhdat[t]), image_clear(im),
                        method=('nearest' if t == 'varea' else 'linear'),
                        address=addr, dtype=dtyp)
                    note('    - Exporting volume file: %s' % flnm)
                    nyio.save(flnm, vol)
                else:
                    note('    - Not overwriting existing file: %s' % flnm)
        note('   Subject %s finished!' % sub.name)
    return 0




####################################################################################################
# cmag.py
# Cortical magnification caclculation code and utilities.
# by Noah C. Benson

import os, sys, six, pimms
import numpy            as np
import scipy            as sp
import scipy.sparse     as sps
import numpy.linalg     as npla
import pyrsistent       as pyr

from .. import geometry as geo
from .. import mri      as mri

from ..util          import (zinv, simplex_summation_matrix, curry, to_hemi_str, flattest, zdivide)
from .retinotopy     import (extract_retinotopy_argument, retinotopy_data, as_retinotopy,
                             retinotopic_field_sign)
from ..geometry.util import (line_segment_intersection_2D, cartesian_to_barycentric_2D)

def mag_data(hemi, retinotopy='any', surface='midgray', mask=None,
             weights=Ellipsis, weight_min=0, weight_transform=Ellipsis,
             visual_area=None, visual_area_mask=Ellipsis,
             eccentricity_range=None, polar_angle_range=None):
    '''
    mag_data(hemi) yields a map of visual/cortical magnification data for the given hemisphere.
    mag_data(mesh) uses the given mesh.
    mag_data([arg1, arg2...]) maps over the given hemisphere or mesh arguments.
    mag_data(subject) is equivalent to mag_data([subject.lh, subject.rh]).
    mag_data(mdata) for a valid magnification data map mdata (i.e., is_mag_data(mdata) is True or
      mdata is a lazy map with integer keys) always yields mdata without considering any additional
      arguments.

    The data structure returned by magdata is a lazy map containing the keys:
      * 'surface_coordinates': a (2 x N) or (3 x N) matrix of the mesh coordinates in the mask
        (usually in mm).
      * 'visual_coordinates': a (2 x N) matrix of the (x,y) visual field coordinates (in degrees).
      * 'surface_areas': a length N vector of the surface areas of the faces in the mesh.
      * 'visual_areas': a length N vector of the areas of the faces in the visual field.
      * 'mesh': the full mesh from which the surface coordinates are obtained.
      * 'submesh': the submesh of mesh of just the vertices in the mask (may be identical to mesh).
      * 'mask': the mask used.
      * 'retinotopy_data': the full set of retinotopy_data from the hemi/mesh; note that this will
        include the key 'weights' of the weights actually used and 'visual_area' of the found or
        specified visual area.
      * 'masked_data': the subsampled retinotopy data from the hemi/mesh.
    Note that if a visual_area property is found or provided (see options below), instead of
    yielding a map of the above, a lazy map whose keys are the visual areas and whose values are the
    maps described above is yielded instead.

    The following named options are accepted (in order):
      * retinotopy ('any') specifies the value passed to the retinotopy_data function to obtain the
        retinotopic mapping data; this may be a map of such data.
      * surface ('midgray') specifies the surface to use.
      * mask (None) specifies the mask to use.
      * weights, weight_min, weight_transform (Ellipsis, 0, Ellipsis) are used as in the
        to_property() function  in neuropythy.geometry except weights, which, if equal to Ellipsis,
        attempts to use the weights found by retinotopy_data() if any.
      * visual_area (Ellipsis) specifies the property to use for the visual area label; Ellipsis is
        equivalent to whatever visual area label is found by the retinotopy_data() function if any.
      * visual_area_mask (Ellipsis) specifies which visual areas to include in the returned maps,
        assuming a visual_area property is found; Ellipsis is equivalent to everything but 0; None
        is equivalent to everything.
      * eccentricity_range (None) specifies the eccentricity range to include.
      * polar_angle_range (None) specifies the polar_angle_range to include.
    '''
    if is_mag_data(hemi): return hemi
    elif pimms.is_lazy_map(hemi) and pimms.is_vector(hemi.keys(), 'int'): return hemi
    if mri.is_subject(hemi): hemi = (hemi.lh. hemi.rh)
    if pimms.is_vector(hemi):
        return tuple([mag_data(h, retinotopy=retinotopy, surface=surface, mask=mask,
                               weights=weights, weight_min=weight_min,
                               weight_transform=weight_transform, visual_area=visual_area,
                               visual_area_mask=visual_area_mask,
                               eccentricity_range=eccentricity_range,
                               polar_angle_range=polar_angle_range)
                      for h in hemi])
    # get the mesh
    mesh = geo.to_mesh((hemi, surface))
    # First, find the retino data
    retino = retinotopy_data(hemi, retinotopy)
    retino = dict(retino)
    # we can process the rest the mask now, including weights and ranges
    if weights is Ellipsis: weights = retino.get('variance_explained', None)
    mask = hemi.indices if mask is None else hemi.mask(mask, indices=True)
    (arng,erng) = (polar_angle_range, eccentricity_range)
    (ang,ecc) = (retino['polar_angle'], retino['eccentricity'])
    if pimms.is_str(arng):
        tmp = to_hemi_str(arng)
        arng = (-180,0) if tmp == 'rh' else (0,180) if tmp == 'lh' else (-180,180)
    elif arng is None:
        tmp = ang[mask]
        tmp = tmp[np.isfinite(tmp)]
        arng = (np.min(tmp), np.max(tmp))
    if erng is None:
        tmp = ecc[mask]
        tmp = tmp[np.isfinite(tmp)]
        erng = (0, np.max(tmp))
    elif pimms.is_scalar(erng): erng = (0, erng)
    (ang,wgt) = hemi.property(retino['polar_angle'], weights=weights, weight_min=weight_min,
                              weight_transform=weight_transform, yield_weight=True)
    ecc = hemi.property(retino['eccentricity'], weights=weights, weight_min=weight_min,
                        weight_transform=weight_transform, data_range=erng)
    # apply angle range if given
    ((mn,mx),mid) = (arng, np.mean(arng))
    oks = mask[np.isfinite(ang[mask])]
    u = ang[oks]
    u = np.mod(u + 180 - mid, 360) - 180 + mid
    ang[oks[np.where((mn <= u) & (u < mx))[0]]] = np.inf
    # mark/unify the out-of-range ones
    bad = np.where(np.isinf(ang) | np.isinf(ecc))[0]
    ang[bad] = np.inf
    ecc[bad] = np.inf
    wgt[bad] = 0
    wgt *= zinv(np.sum(wgt[mask]))
    # get visual and surface coords
    vcoords = np.asarray(as_retinotopy(retino, 'geographical'))
    scoords = mesh.coordinates
    # now figure out the visual area so we can call down if we need to
    if visual_area is Ellipsis: visual_area = retino.get('visual_area', None)
    if visual_area is not None: retino['visual_area'] = visual_area
    if wgt is not None: retino['weights'] = wgt
    rdata = pimms.persist(retino)
    # calculate the range area
    (tmn,tmx) = [np.pi/180.0 * u for u in arng]
    if tmx - tmn >= 2*np.pi: (tmn,tmx) = (-np.pi,np.pi)
    (emn,emx) = erng
    rarea = 0.5 * (emx*emx - emn*emn) * (tmx - tmn)
    # okay, we have the data organized; we can do the calculation based on this, but we may have a
    # visual area mask to apply as well; here's how we do it regardless of mask
    def finish_mag_data(mask):
        if len(mask) == 0: return None
        # now that we have the mask, we can subsample
        if np.array_equal(mask, mesh.indices): submesh = mesh
        else:                                  submesh = mesh.submesh(mesh.labels[mask])
        mask = mesh.tess.index(submesh.labels)
        mdata = pyr.pmap({k:(v[mask]   if pimms.is_vector(v) else
                             v[:,mask] if pimms.is_matrix(v) else
                             None)
                          for (k,v) in six.iteritems(rdata)})
        fs = submesh.tess.indexed_faces
        (vx, sx)  = [x[:,mask]                        for x in (vcoords, scoords)]
        (vfx,sfx) = [np.asarray([x[:,f] for f in fs]) for x in (vx,      sx)]
        (va, sa)  = [geo.triangle_area(*x)            for x in (vfx, sfx)]
        return pyr.m(surface_coordinates=sx, visual_coordinates=vx,
                     surface_areas=sa,       visual_areas=va,
                     mesh=mesh,              submesh=submesh,
                     retinotopy_data=rdata,  masked_data=mdata,
                     mask=mask,              area_of_range=rarea)
    # if there's no visal area, we just use the mask as is
    if visual_area is None: return finish_mag_data(mask)
    # otherwise, we return a lazy map of the visual area mask values
    visual_area = hemi.property(visual_area, mask=mask, null=0, dtype=np.int)
    vam = (np.unique(visual_area)                    if visual_area_mask is None     else
           np.setdiff1d(np.unique(visual_area), [0]) if visual_area_mask is Ellipsis else
           np.unique(list(visual_area_mask)))
    return pimms.lazy_map({va: curry(finish_mag_data, mask[visual_area[mask] == va])
                           for va in vam})
def is_mag_data(mdat):
    '''
    is_mag_data(dat) yields True if the given data is a valid set of magnification data and False
      otherwise.

    Note that this does not return True for all valid return values of the mag_data() function:
    specifically, if the mag_data() function yields a list of mag-data maps or a lazy-map of the
    mag-data maps split out by visual area, then this will return False. This function only returns
    True for a map of mag data itself.
    '''
    if not pimms.is_map(mdat): return False
    for k in ['surface_coordinates', 'visual_coordinates', 'mesh', 'submesh', 'mask',
              'retinotopy_data', 'masked_data', 'surface_areas', 'visual_areas']:
        if k not in mdat: return False
    return True

def parse_toopt_facevertex(to):
    '''
    Parses the `to` optional-argument of the functions below; returns either 'vertex' or 'face',
    or raises an error.
    '''
    if to in [None,Ellipsis]: return 'vertex'
    if not pimms.is_str(to): raise ValueError('could not parse `to` argument: %s' % (to,))
    to = to.lower()
    if to in ['vertices','vertex','nodes','node','points','v','vtx','vtcs','pts']: return 'vertex'
    elif to in ['faces','f','triangles','t','tri','tris','face','triangle']: return 'face'
    else: raise ValueError('could not parse `to` argument: %s' % (to,))
    
def rtmag_potential(submesh, X0, mask=Ellipsis, fieldsign=None):
    '''
    rtmag_potential(mesh, viscoords, ...) yields the
      radial/tangential cortical magnification term of the potential field.

    This should generally not be called directly and instead should be obtained from the
    clean_retinotopy_potential() function instead.
    '''
    import neuropythy.optimize as op
    if fieldsign == 0: fieldsign = None
    # A few other handy pieces of data we can extract:
    sxyz = submesh.coordinates
    n = submesh.vertex_count
    (u,v) = submesh.tess.indexed_edges
    selen = submesh.edge_lengths
    sarea = submesh.face_areas
    m = submesh.tess.edge_count
    fs = submesh.tess.indexed_faces
    neis = submesh.tess.indexed_neighborhoods
    fangs = submesh.face_angles
    # we're adding r and t (radial and tangential visual magnification) pseudo-parameters to
    # each vertex; r and t are derived from the position of other vertices; our first step is
    # to derive these values; for this we start with the parameters themselves:
    (x,y) = [op.identity[np.arange(k, 2*n, 2)] for k in (0,1)]
    # okay, we need to setup a bunch of least-squares solutions, one for each vertex:
    nneis = np.asarray([len(nn) for nn in neis])
    maxneis = np.max(nneis)
    thts = op.atan2(y, x)
    eccs = op.compose(op.piecewise(op.identity, ((-1e-9, 1e-9), 1)),
                      op.sqrt(x**2 + y**2))
    coss = x/eccs
    sins = y/eccs
    # organize neighbors:
    # neis becomes a list of rows of 1st neighbor, second neighbor etc. with -1 indicating none
    neis = np.transpose([nei + (-1,)*(maxneis - len(nei)) for nei in neis])
    qnei = (neis > -1) # mark where there are actually neighbors
    neis[~qnei] = 0 # we want the -1s (now 0s) to behave okay when passed to a potential index
    # okay, walk through the neighbors setting up the least squares
    (r, t) = (None, None)
    for (k,q,nei) in zip(range(len(neis)), qnei.astype('float'), neis):
        xx = x[nei] - x
        yy = y[nei] - y
        sd = np.sum((sxyz[:,nei].T - sxyz[:,k])**2, axis=1)
        (xx, yy) = (xx*coss + yy*sins, yy*coss - xx*sins)
        xterm = (op.abs(xx) * q)
        yterm = (op.abs(yy) * q)
        r = xterm if r is None else (r + xterm)
        t = yterm if t is None else (t + yterm)
    (r, t) = [uu * zinv(nneis) for uu in (r, t)]
    # for neighboring edges, we want r and t to be similar to each other
    f_rtsmooth = op.sum((r[v]-r[u])**2 + (t[v]-t[u])**2) / m
    # we also want r and t to predict the radial and tangential magnification of the node, so
    # we want to make sure that edges are the right distances away from each other based on the
    # surface edge lengths and the distance around the vertex at the center
    # for this we'll want some constant info about the surface edges/angles
    # okay, in terms of the visual field coordinates of the parameters, we will want to know
    # the angular position of each node
    # organize face info
    mnden   = 0.0001
    (e,qs,qt) = np.transpose([(i,e[0],e[1]) for (i,e) in enumerate(submesh.tess.edge_faces)
                              if len(e) == 2 and selen[i] > mnden
                              if sarea[e[0]] > mnden and sarea[e[1]] > mnden])
    (fis,q) = np.unique(np.concatenate([qs,qt]), return_inverse=True)
    (qs,qt)   = np.reshape(q, (2,-1))
    o       = len(fis)
    faces   = fs[:,fis]
    fangs   = fangs[:,fis]
    varea   = op.signed_face_areas(faces)
    srfangmtx = sps.csr_matrix(
        (fangs.flatten(),
         (faces.flatten(), np.concatenate([np.arange(o), np.arange(o), np.arange(o)]))),
        (n, o))
    srfangtot = flattest(srfangmtx.sum(axis=1))
    # normalize this angle matrix by the total and put it back in the same order as faces
    srfangmtx = zdivide(srfangmtx, srfangtot / (np.pi*2)).tocsr().T
    nrmsrfang = np.array([sps.find(srfangmtx[k])[2][np.argsort(fs[:,k])] for k in range(o)]).T
    # okay, now compare these to the actual angles;
    # we also want to know, for each edge, the angle relative to the radial axis; let's start
    # by organizing the faces into the units we compute over:
    (fa,fb,fc) = [np.concatenate([faces[k], faces[(k+1)%3], faces[(k+2)%3]]) for k in range(3)]
    atht = thts[fa]
    # we only have to worry about the (a,b) and (a,c) edges now; from the perspective of a...
    bphi = op.atan2(y[fb] - y[fa], x[fb] - x[fa]) - atht
    cphi = op.atan2(y[fc] - y[fa], x[fc] - x[fa]) - atht
    ((bcos,bsin),(ccos,csin)) = bccssn = [(op.cos(q),op.sin(q)) for q in (bphi,cphi)]
    # the distance should be predicted by surface edge length times ellipse-magnification
    # prediction; we have made uphi and vphi so that radial axis is x axis and tan axis is y
    (ra,ta) = (op.abs(r[fa]), op.abs(t[fa]))
    bslen = np.sqrt(np.sum((sxyz[:,fb] - sxyz[:,fa])**2, axis=0))
    cslen = np.sqrt(np.sum((sxyz[:,fc] - sxyz[:,fa])**2, axis=0))
    bpre_x = bcos * ra * bslen
    bpre_y = bsin * ta * bslen
    cpre_x = ccos * ra * cslen
    cpre_y = csin * ta * cslen
    # if there's a global field sign, we want to invert these predictions when the measured
    # angle is the wrong sign
    if fieldsign is not None:
        varea_f = varea[np.concatenate([np.arange(o) for _ in range(3)])] * fieldsign
        fspos = (op.sign(varea_f) + 1)/2
        fsneg = 1 - fspos
        (bpre_x,bpre_y,cpre_x,cpre_y) = (
            bpre_x*fspos - cpre_x*fsneg, bpre_y*fspos - cpre_y*fsneg,
            cpre_x*fspos - bpre_x*fsneg, cpre_y*fspos - bpre_y*fsneg)
    (ax,ay,bx,by,cx,cy) = [x[fa],y[fa],x[fb],y[fb],x[fc],y[fc]]
    (cost,sint) = [op.cos(atht), op.sin(atht)]
    (bpre_x, bpre_y) = (bpre_x*cost - bpre_y*sint + ax, bpre_x*sint + bpre_y*cost + ay)
    (cpre_x, cpre_y) = (cpre_x*cost - cpre_y*sint + ax, cpre_x*sint + cpre_y*cost + ay)
    # okay, we can compare the positions now...
    f_rt = op.sum((bpre_x-bx)**2 + (bpre_y-by)**2 + (cpre_x-cx)**2 + (cpre_y-cy)**2) * 0.5/o
    f_vmag = f_rtsmooth # + f_rt #TODO: the rt part of this needs to be debugged
    object.__setattr__(f_vmag, 'meta_data', pyr.m(f_rtsmooth=f_rtsmooth, f_rt=f_rt))
    return f_vmag

def disk_vmag(hemi, retinotopy='any', yields='axes', min_cod=0, npoints=50, **kw):
    '''
    disk_vmag(mesh) yields the visual magnification based on the projection of disks on the cortical
      surface into the visual field.

    All options accepted by mag_data() are accepted by disk_vmag(). In addition, the parameters
    yields and min_cod may be provided. The min_cod parameter indicates the minimum coefficient of
    determination (r-squared), calculated between the fitted-ellipse and the vetex neighbor's raw
    positions, that is needed to be included in the returrn values. The yields option determines 
    what the return value should be. The default value is 'axes', but the following values are
    accepted:
      * 'axes' indicates that the return value should be an (n x 2 x 2) array where n is the number
        of vertices in the mesh or cortex; each 2x2 matrix is the [rad_x rad_y; tan_x tan_y] axes.
      * 'cod' indicates that only the coefficient of determination for the least-squares fit should
        be returned.
      * 'all' indicates that the return value should be (axes, cod).
    In all cases, nans indicate vertices that were not part of the retinotopy mask, that did not
    have CODs above the threshold, or that had too few neighbors to fit an ellipse.
    '''
    mdat = mag_data(hemi, retinotopy=retinotopy, **kw)
    if pimms.is_vector(mdat): return tuple([face_vmag(m, to=to) for m in mdat])
    elif pimms.is_vector(mdat.keys(), 'int'):
        return pimms.lazy_map({k: curry(lambda k: disk_vmag(mdat[k], to=to), k)
                               for k in six.iterkeys(mdat)})
    # for disk cmag we start by making sets of circular points around each vertex
    msh  = mdat['submesh']
    n    = msh.vertex_count
    N    = hemi.vertex_count
    vxy  = mdat['visual_coordinates'].T
    sxy  = msh.coordinates.T
    parts = msh.tess.neighborhood_face_partition
    emax = len(parts)
    mindist = np.full(n, np.inf)
    angtot = np.zeros(n)
    angs = []
    for (a,b,c) in parts:
        abvec = (sxy[b] - sxy[a])
        acvec = (sxy[c] - sxy[a])
        ablen = np.sqrt(np.sum(abvec**2, axis=1))
        aclen = np.sqrt(np.sum(acvec**2, axis=1))
        mindist[a] = np.min([mindist[a], ablen, aclen], axis=0)
        abvec /= ablen[:,None]
        acvec /= aclen[:,None]
        angs.append(np.arccos(np.sum(abvec*acvec, axis=1)))
        angtot[a] += angs[-1]
    ellipses = np.full((n,npoints,2), np.nan)
    angprog = np.zeros(n)
    ang_per_pt = angtot / npoints
    for (ang,part) in zip(angs, parts):
        (a,b,c) = part
        sabvec = (sxy[b] - sxy[a])
        sacvec = (sxy[c] - sxy[a])
        sablen = np.sqrt(np.sum(sabvec**2, axis=1))
        saclen = np.sqrt(np.sum(sacvec**2, axis=1))
        vabvec = (vxy[b] - vxy[a])
        vacvec = (vxy[c] - vxy[a])
        vablen = np.sqrt(np.sum(vabvec**2, axis=1))
        vaclen = np.sqrt(np.sum(vacvec**2, axis=1))
        uabvec = vabvec * (mindist[a] / sablen)[:,None]
        uacvec = vacvec * (mindist[a] / saclen)[:,None]
        uablen = np.sqrt(np.sum(uabvec**2, axis=1))
        uaclen = np.sqrt(np.sum(uacvec**2, axis=1))
        uabtht = np.arctan2(uabvec[:,1], uabvec[:,0])
        uactht = np.arctan2(uacvec[:,1], uacvec[:,0])
        angprog_start = np.array(angprog[a])
        p0 = np.round((angprog[a] / angtot[a]) * npoints).astype('int')
        ii = np.arange(len(a))
        aa = a
        while True:
            angprog_aa = angprog[aa]
            ii = ii[(angprog_aa - angprog_start[ii] < ang[ii]) &
                    (angprog_aa + ang_per_pt[aa] <= angtot[aa])]
            (aa,bb,cc) = (a[ii], b[ii], c[ii])
            if len(ii) == 0: break
            (rb,rc,thb,thc) = (uablen[ii], uaclen[ii], uabtht[ii], uactht[ii])
            angw = (angprog[aa] - angprog_start[ii]) / ang[ii]
            iangw = 1 - angw
            tht = iangw*thb + angw*thc
            r = iangw*rb + angw*rc
            atpt = np.round(angprog[aa] / angtot[aa] * npoints).astype('int')
            ellipses[aa, atpt, :] = np.transpose([np.cos(tht)*r, np.sin(tht)*r])
            angprog[aa] += ang_per_pt[aa]
    # At this point, ellipses is the set of ellipse points to fit for each neighborhood.
    # Its dimensions repesent (vertices, neighbors, x/y coords); we want to rotate the
    # points to be along their center's implied rad/tan axis.
    vrs  = np.sqrt(np.sum(vxy**2, axis=1))
    irs  = zinv(vrs)
    coss = vxy[:,0] * irs
    sins = vxy[:,1] * irs
    # rotating each ellipse by negative-theta gives us x=radial and y=tangential
    cels = (coss * ellipses.T)
    sels = (sins * ellipses.T)
    rots = np.transpose([cels[0] + sels[1], cels[1] - sels[0]], [1,2,0])
    rsrt = np.sqrt(np.sum(rots**2, axis=2)).T
    (csrt,snrt) = zinv(rsrt) * rots.T
    # At this point, csrt and snrt have dimensions that represent (vertices, neighbors)
    # ... (a*cos(rots))**2 + (b*sin(rots))**2 ~= r(rots) where a = radial vmag and b = tan vmag
    axes = []
    cods = []
    idxs = []
    for (r,c,s,irad,i) in zip(rsrt,csrt,snrt,zinv(mindist),range(n)):
        # if the center point is way outside the min/max, skip it
        (x,y) = (r*c, r*s)
        if len(np.unique(np.sign(x))) < 2 or len(np.unique(np.sign(y))) < 2: continue
        mudst = np.sqrt(np.sum(np.mean([x, y], axis=1)**2))
        #if mudst > np.min(r): continue
        # okay, fit an ellipse...
        fs = np.transpose([c,s])**2
        try:
            (ab,rss,rnk,svs) = np.linalg.lstsq(fs, r**2, rcond=None)
            # ab = np.sqrt(np.abs(ab)) # Why the sqrt?
            ab = np.abs(ab)
            if len(rss) == 0 or rnk < 2: continue # or np.min(svs/np.sum(svs)) < 0.01: continue
            cod = 1 - rss[0]*zinv(np.sum(r**2))
            if cod < min_cod: continue
            axes.append(ab * irad)
            cods.append(cod)
            idxs.append(i)
        except Exception as e: continue
    (axes,cods,idxs) = [np.asarray(u) for u in (axes, cods, idxs)]
    # make the return value; this is a 2x2 matrix for each vertex
    if yields != 'cod':
        raxes = np.full((N, 2), np.nan)
        raxes[idxs,:] = axes
        if yields == 'axes': return raxes
    rcods = np.full(N, np.nan)
    rcods[idxs] = cods
    if yields == 'cod': return rcods
    else: return (raxes, rcods)

def face_vmag(hemi, retinotopy='any', to=None, null=np.nan, **kw):
    '''
    face_vmag(mesh) yields the visual magnification based on the projection of individual faces on
      the cortical surface into the visual field.
    face_vmag(mdat) uses the given magnification data mdat (as returned from mag_data()); if valid
      magnification data is passed then all options related to the mag_data() function are ignored.

    All options accepted by mag_data() are accepted by face_vmag().

    The additional optional arguments are also accepted:
      * to (default: None) specifies that the resulting data should be transformed in some way;
        these transformations are:
          * 'faces': returns a property of the visual magnification value of each face;
          * 'vertices': returns a property of the visual magnification value of each vertex, as
            determined by averaging the magnification 
          * None (the detault) is equivalent to 'faces'
      * null (default: nan) specifies the value to use in the return-value for faces or
        vertices outside the mesh.
    '''
    to = parse_toopt_facevertex(to)
    if pimms.is_vector(hemi):
        return tuple([face_vmag(m, to=to) for m in mdat])
    if not is_mag_data(hemi):
        if pimms.is_map(hemi) and all(pimms.is_int(k) for k in six.iterkeys(hemi)):
            m = {k: curry(lambda k: face_vmag(mdat[k], to=to), k) for k in ks}
            return pimms.lmap(m)
        else:
           hemi = mag_data(hemi, retinotopy=retinotopy, **kw)
    mdat = hemi
    # Okay, we have a single mag-data dict; get the face visual and cortical areas:
    mesh = mdat['mesh']
    submesh = mdat['submesh']
    vismesh = submesh.copy(coordinates=mdat['visual_coordinates'])
    subface_vmags = zdivide(vismesh.face_areas, submesh.face_areas, null=0)
    face_vmags = np.full(mesh.tess.face_count, null)
    face_vmags[mesh.tess.index[submesh.tess.faces]] = subface_vmags
    # convert these if the to argument so-requires it:
    if to == 'face': return face_vmags
    # possible #TODO here: super-tesselate the mesh and calculate the actual visual vertex areas
    # then divide by the vertex cortical surface ares to do this properly.
    vmag = np.full(mesh.vertex_count, null)
    for (vtx,faces) in six.iteritems(mdat['submesh'].tess.vertex_face_index):
        tmp = face_vmags[list(faces)]
        ii = mesh.tess.index(vtx)
        jj = np.where(np.isfinite(tmp))[0]
        jj = jj[tmp[jj] > 0]
        if len(jj) > 0: vmag[ii] = np.mean(tmp[jj])
    return vmag

def face_rtcmag(hemi, retinotopy='any', to=None, **kw):
    '''
    face_rtcmag(mesh) yields the cortical magnification matrices of each face in
      the given mesh based on the projection of individual faces on the
      cortical surface into the visual field.
    face_rtcmag(mdat) uses the given magnification data mdat (as returned from
      mag_data()); if valid magnification data is passed then all options
      related to the mag_data() function are ignored.

    The return value of face_rtcmag() is an tuple of (rad_cmag, tan_cmag, theta)
    where rad_cmag and tan_cmag are both linear cortical magnifications in the
    associated directions, and theta is the angle between them on cortex.

    All options accepted by mag_data() are accepted by face_rtcmag().

    The additional optional arguments are also accepted:
      * to (default: None) specifies that the resulting data should be
        transformed in some way; these transformations are:
          * None or 'data': returns the full magnification data without 
            transformation;
          * 'faces': returns a property of the visual magnification value of
            each face;
          * 'vertices': returns a property of the visual magnification value of
            each vertex, as determined by averaging the magnification.
    '''
    if pimms.is_vector(hemi):
        return tuple([face_rtcmag(m, to=to) for m in mdat])
    if not is_mag_data(hemi):
        if (pimms.is_map(hemi) and 
            all(pimms.is_int(k) for k in six.iterkeys(hemi))):
            m = {k: curry(lambda k: face_rtcmag(mdat[k], to=to), k) for k in ks}
            return pimms.lmap(m)
        else:
            return face_rtcmag(mag_data(hemi, retinotopy=retinotopy, **kw), to=to)
    to = parse_toopt_facevertex(to)
    mdat = hemi
    # Okay, we have a single mag-data dict; get the face visual and cortical areas:
    mesh = mdat['mesh']
    submesh = mdat['submesh']
    vismesh = submesh.copy(coordinates=mdat['visual_coordinates'])
    face_vmags = zdivide(vismesh.face_areas, submesh.face_areas, null=0)
    # To determine radial and tangential, we start by drawing lines from each 
    # face center to whichever other triangle side in the radial and tangential
    # directions.
    fx0 = vismesh.face_centers
    (ax,bx,cx) = abcx = vismesh.face_coordinates
    ecc = np.sqrt(np.sum(fx0**2, axis=0))
    tht = np.arctan2(fx0[1], fx0[0])
    u_rad = fx0 * zinv(ecc, null=np.nan)
    u_tan = np.array([-u_rad[1], u_rad[0]])
    # Get the intersection of these u_rad and u_tan with the triangle sides
    radlines = [np.hstack([u,u,u]) for u in (fx0, fx0+u_rad)]
    tanlines = [np.hstack([u,u,u]) for u in (fx0, fx0+u_tan)]
    trisides = [np.hstack([ax,bx,cx]), np.hstack([bx,cx,ax])]
    isect_rad = line_segment_intersection_2D(radlines, trisides)
    isect_tan = line_segment_intersection_2D(tanlines, trisides)
    vis_segs = []
    for (isect_dat, u_dat) in zip([isect_rad, isect_tan], [u_rad, u_tan]):
        isects_wnan = np.array([u.T for u in np.reshape(np.transpose(isect_dat), (3,-1,2))])
        # There should be 2 intersections per column of these matrices.
        isects = np.zeros((2,) + isects_wnan.shape[1:])
        isects_fin = np.isfinite(isects_wnan[:,0,:])
        nisects = np.sum(isects_fin, axis=0)
        ii3 = (nisects == 3)
        if np.any(ii3):
            wh3 = np.where(ii3)[0]
            (iab, ibc, ica) = isects_wnan
            d2b = np.sum((iab[:,wh3] - ibc[:,wh3])**2, axis=0)
            d2c = np.sum((ibc[:,wh3] - ica[:,wh3])**2, axis=0)
            d2a = np.sum((ica[:,wh3] - iab[:,wh3])**2, axis=0)
            dists = [d2b, d2c, d2a]
            far_isects = [ica, iab, ibc]
            small = np.argmin(dists, axis=0)
            # The two with the smallest distance between them we average
            for s in (0,1,2):
                iis = np.where(small == s)[0]
                (far,cl1,cl2) = [far_isects[(s+k)%3][:, wh3[iis]] for k in (0,1,2)]
                cl = np.mean([cl1,cl2], axis=0)
                isects[0,:,wh3[iis]] = cl.T
                isects[1,:,wh3[iis]] = far.T
        # For the rest, we can just take the good two intersections.
        ii2 = ~ii3
        k = np.argmin(isects_fin, axis=0)[ii2]
        rng = np.arange(isects_wnan.shape[2])[ii2]
        isects[0,:,ii2] = isects_wnan[(k+1)%3, :, rng]
        isects[1,:,ii2] = isects_wnan[(k+2)%3, :, rng]
        # For all the intersections, we want to make sure that they are oriented
        # correctly w.r.t rad/tan (i.e., make we didn't flip the vectors).
        ivec = isects[1] - isects[0]
        sgn = np.sign(np.sum(ivec * u_dat, axis=0))
        ii = np.where(sgn == -1)[0]
        tmp = np.array(isects[1,:,ii])
        isects[1,:,ii] = isects[0,:,ii]
        isects[0,:,ii] = tmp
        vis_segs.append(isects)
    # That gives us all the intersections in visual space.
    # We now want them in cortical space as well, and we want to convert both
    # into distances.
    vis_segs = np.array(vis_segs)
    bc_coords = np.array(
        [[(a, b, 1.0 - a - b)
          for isect in isects
          for (a,b) in [cartesian_to_barycentric_2D(abcx, isect)]]
         for isects in vis_segs])
    srf_abcx = submesh.face_coordinates
    srf_segs = np.array(
        [[np.sum(srf_abcx * isect[:,None,:], axis=0)
          for isect in isects]
         for isects in bc_coords])
    # Okay, convert the segments into lengths.
    srf_rtlen = np.array([np.sqrt(np.sum((a - b)**2, axis=0)) for (a,b) in srf_segs])
    vis_rtlen = np.array([np.sqrt(np.sum((a - b)**2, axis=0)) for (a,b) in vis_segs])
    (rmag, tmag) = srf_rtlen / vis_rtlen
    # Those magnifications fail in consider the angle between the surface lines,
    # though; to correct for this, we just find it first:
    srf_vecs = np.array([b - a for (a,b) in srf_segs])
    srf_vecs *= zinv(np.sqrt(np.sum(srf_vecs**2, axis=1)))[:,None,:]
    theta = np.arccos(np.sum(srf_vecs[0] * srf_vecs[1], axis=0))
    # We also want to preserve +/- angle values to indicate whether there was a
    # flip (positive for positive fieldsign and negative for negative).
    fsign = retinotopic_field_sign(mesh, 'faces', retinotopy=mdat['retinotopy_data'])
    theta *= fsign[mesh.tess.index[submesh.tess.faces]]
    # Now accumulate and return.
    res = []
    ii = mesh.tess.index[submesh.tess.faces]
    for x in (rmag,tmag,theta):
        y = np.full(mesh.tess.face_count, np.nan)
        y[ii] = x
        res.append(y)
    return tuple(res)

@pimms.immutable
class FieldOfView(object):
    '''
    FieldOfView is a class that represents and calculates the field of view in a cortical area.
    '''
    def __init__(self, angle, eccen, sigma,
                 scale=None, weight=None, search_scale=3.0, bins=6,
                 normalize_weights=True, weights_method='height'):
        self.polar_angle = angle
        self.eccentricity = eccen
        self.sigma = sigma
        self.weight = weight
        self.bins = bins
        self.search_scale = search_scale
        self.scale = scale
        self.normalize_weights = normalize_weights
        self.weights_method = weights_method
    @pimms.param
    def polar_angle(pa):  return pimms.imm_array(pa)
    @pimms.param
    def eccentricity(ec): return pimms.imm_array(ec)
    @pimms.param
    def sigma(r):         return pimms.imm_array(r)
    @pimms.param
    def scale(s): return s
    @pimms.param
    def normalize_weights(nw): return bool(nw)
    @pimms.param
    def weights_method(wm):
        if wm is None or wm is Ellipsis: return 'height'
        wm = wm.lower()
        if wm in ['height', 'h', 'z']: return 'height'
        elif wm in ['volume', 'vol', 'v']: return 'volume'
        else: raise ValueError('unrecognized weights_method: %s' % (wm,))
    @pimms.param
    def weight(w):
        if w is None: return None
        w = np.array(w)
        w.setflags(write=False)
        return w
    @pimms.param
    def bins(h): return h
    @pimms.param
    def search_scale(s): return s
    @pimms.value
    def theta(polar_angle):
        return pimms.imm_array(np.mod(np.pi/180.0*(90 - polar_angle) + np.pi, 2*np.pi) - np.pi)
    @pimms.value
    def coordinates(theta, eccentricity):
        x = eccentricity * np.cos(theta)
        y = eccentricity * np.sin(theta)
        return pimms.imm_array(np.transpose([x,y]))
    @pimms.value
    def _weight(weight, polar_angle, normalize_weights, sigma, weights_method):
        if weight is None:
            weight = np.ones(len(polar_angle))
            normalize_weights = True
        if normalize_weights:
            weight = weight / np.sum(weight)
        if weights_method == 'volume':
            weight = weight / (2 * np.pi * sigma)
        return pimms.imm_array(weight)
    @pimms.value
    def sigma_bin_walls(sigma, bins):
        import scipy, scipy.cluster, scipy.cluster.vq as vq
        std = np.std(sigma)
        if np.isclose(std, 0): return pimms.imm_array([0, np.max(sigma)])
        cl = sorted(std * vq.kmeans(sigma/std, bins)[0])
        cl = np.mean([cl[:-1],cl[1:]], axis=0)
        return pimms.imm_array(np.concatenate(([0], cl, [np.max(sigma)])))
    @pimms.value
    def sigma_bins(sigma, sigma_bin_walls):
        bins = []
        for (mn,mx) in zip(sigma_bin_walls[:-1], sigma_bin_walls[1:]):
            ii = np.logical_and(mn < sigma, sigma <= mx)
            bins.append(pimms.imm_array(np.where(ii)[0]))
        return tuple(bins)
    @pimms.value
    def bin_query_distances(sigma_bins, sigma, search_scale):
        return tuple([np.max(sigma[ii])*search_scale for ii in sigma_bins])
    @pimms.value
    def spatial_hashes(coordinates, sigma_bins):
        import scipy, scipy.spatial
        try:              from scipy.spatial import cKDTree as shash
        except Exception: from scipy.spatial import KDTree  as shash
        return tuple([shash(coordinates[ii]) for ii in sigma_bins])
    # Methods
    def __call__(self, x, y=None):
        if y is not None: x = (x,y)
        x = np.asarray(x)
        if len(x.shape) == 1: return self([x])[0]
        x = np.transpose(x) if x.shape[0] == 2 else x
        if not x.flags['WRITEABLE']: x = np.array(x)
        crd = self.coordinates
        sig = self.sigma
        wts = self._weight
        res = np.zeros(x.shape[0])
        for (sh, qd, bi) in zip(self.spatial_hashes, self.bin_query_distances, self.sigma_bins):
            neis = sh.query_ball_point(x, qd)
            res += [
                np.sum(w * np.exp(-0.5 * d2/s**2))
                for (ni,pt) in zip(neis,x)
                for ii in [bi[ni]]
                for (w,s,d2) in [(wts[ii], sig[ii], np.sum((crd[ii] - pt)**2, axis=1))]]
        return res
def field_of_view(mesh, retinotopy='any', mask=None, search_scale=3.0, bins=6, weights=Ellipsis,
                  normalize_weights=True, weights_method='height'):
    '''
    field_of_view(obj) yields a field-of-view function for the given vertex-set object or mapping of
      retinotopy data obj.

    The field-of-view function is a measurement of how much total pRF weight there is at each point
    in the visual field; essentially it is the sum of all visual-field Gaussian pRFs, where the
    Gaussians are normalized both by the pRF size (i.e., each pRF is a 2D normal distribution whose
    center comes from its polar angle and eccentricity and whose sigma parameter is the pRF radius)
    and the weight (if any weight such as the variance explained is included in the retinotopy data,
    then the normal distributions are multiplied by this weight, otherwise weights are considered to
    be uniform). Note that the weights are normalized by their sum, so the resulting field-of-view
    function should be a valid probability distribution over the visual field. To disable this,
    you can use the nomalize_weights=False option.

    The returned object fov = field_of_view(obj) is a special field-of-view object that can be
    called as fov(x, y) or fov(coord) where x and y may be lists or values and coord may be a 2D
    vector or an (n x 2) matrix.

    The following options may be provided (also in argument order):
      * retinotopy (default: 'any') specifies the retinotopy data to be used; if this is a string,
        then runs retinotopy_data(obj, retinotopy) and uses the result, otherwise runs
        retinotopy_data(retinotopy) and uses the result; note that this must have a radius parameter
        as well as polar angle and eccentricity or equivalent parameters that specify the pRF center
      * mask (default: None) specifies the mask that should be used; this is interpreted by a call
        to mesh.mask(mask)
      * search_scale (default: 3.0) specifies how many Gaussian standard deviations away from a pRF
        center should be included when calculating the sum of Gaussians at a point
      * bins (default: 6) specifies the number of bins to divide the pRFs into based on the radius
        value; this is used to prune the search of pRFs that overlap a point to just those
        reasonanly close to the point in question
      * weights (default: Ellipsis) specifies the weights that should be used. Ellipsis inidicates
        that whatever weights are found in the retinotopy data should be used and none should be
        used otherwise.
      * normalize_weights (default: True) may be set to False to prevent normalization of the
        weights. The default value of True normalizes the weights to be equal to 1.
      * weights_method (default: 'height') specifies whether the weights should be considered the
        heights or the volumes of the Gaussians pRFs.
    '''
    # First, find the retino data
    if pimms.is_str(retinotopy):
        retino = retinotopy_data(mesh, retinotopy)
    else:
        retino = retinotopy_data(retinotopy)
    # Convert from polar angle/eccen to longitude/latitude
    (ang,ecc) = as_retinotopy(retino, 'visual')
    if 'radius' not in retino:
        raise ValueError('Retinotopy data must contain a radius, sigma, or size value')
    sig = retino['radius']
    if weights is Ellipsis or weights is True:
        wgt = next((retino[q] for q in ('variance_explained', 'weight') if q in retino), None)
    elif weights is None or weights is False:
        wgt = None
    else:
        wgt = mesh.property(weights)
    # Get the indices we care about
    ii = geo.to_mask(mesh, mask, indices=True)
    return FieldOfView(ang[ii], ecc[ii], sig[ii], weight=wgt[ii],
                       search_scale=search_scale, bins=bins,
                       normalize_weights=normalize_weights, weights_method=weights_method)

@pimms.immutable
class ArealCorticalMagnification(object):
    '''
    ArealCorticalMagnification is a class that represents and calculates coritcal magnification.
    '''
    def __init__(self, angle, eccen, sarea, weight=None, nnearest=200):
        self.polar_angle = angle
        self.eccentricity = eccen
        self.surface_area = sarea
        self.weight = weight
        self.nnearest = nnearest
    @pimms.param
    def polar_angle(pa):  return pimms.imm_array(pa)
    @pimms.param
    def eccentricity(ec): return pimms.imm_array(ec)
    @pimms.param
    def surface_area(sa): return pimms.imm_array(sa)
    @pimms.param
    def weight(w):
        if w is None: return None
        w = np.array(w)
        w /= np.sum(w)
        w.setflags(write=False)
        return w
    @pimms.param
    def nnearest(s): return int(s)
    @pimms.value
    def theta(polar_angle):
        return pimms.imm_array(np.mod(np.pi/180.0*(90 - polar_angle) + np.pi, 2*np.pi) - np.pi)
    @pimms.value
    def coordinates(theta, eccentricity):
        x = eccentricity * np.cos(theta)
        y = eccentricity * np.sin(theta)
        return pimms.imm_array(np.transpose([x,y]))
    @pimms.value
    def _weight(weight, surface_area):
        if weight is None: weight = np.ones(len(surface_area))
        weight = weight * surface_area
        weight /= np.sum(weight)
        weight.setflags(write=False)
        return weight
    @pimms.value
    def spatial_hash(coordinates):
        import scipy, scipy.spatial
        try:              from scipy.spatial import cKDTree as shash
        except Exception: from scipy.spatial import KDTree  as shash
        return shash(coordinates)
    # Static helper function
    @staticmethod
    def _xy_to_matrix(x, y=None):
        if y is not None: x = (x,y)
        x = np.asarray(x)
        return x if len(x.shape) == 1 or x.shape[1] == 2 else x.T
    # Methods
    def nearest(self, x, y=None, n=Ellipsis):
        n = self.nnearest if n is Ellipsis else n
        x = ArealCorticalMagnification._xy_to_matrix(x,y)
        if not x.flags['WRITEABLE']: x = np.array(x)
        return self.spatial_hash.query(x, n)
    def __call__(self, x, y=None):
        (d,ii) = self.nearest(x, y)
        n = self.spatial_hash.n
        bd = (ii == n)
        ii[bd] = 0
        carea = np.reshape(self.surface_area[ii.flatten()], ii.shape)
        carea[bd] = np.nan
        d[bd] = np.nan
        if len(d.shape) == 1:
            varea = np.pi * np.nanmax(d)**2
            carea = np.nansum(carea)
        else:
            varea = np.pi * np.nanmax(d, axis=1)**2
            carea = np.nansum(carea, axis=1)
        return carea / varea
    def visual_pooling_area(self, x, y=None):
        (d,ii) = self.nearest(x, y)
        return d[-1] if len(d.shape) == 1 else d[:,-1]    
def areal_cmag(mesh, retinotopy='any', mask=None, surface_area=None, weight=None, nnearest=None):
    '''
    areal_cmag(obj) yields an areal cortical magnification function for the given vertex-set object
      or mapping of retinotopy data obj. If obj is a mapping of retinotopy data, then it must also
      contain the key 'surface_area' specifying the surface area of each vertex in mm^2.

    The areal cmag function is a measurement of how much surface area in a retinotopy dataset is
    found within a circle found in the visual field. The size of the circle is chosen automatically
    by looking up the <n> nearest neighbors of the point in question, summing their surface areas,
    and dividing by the area of the circle in the visual field that is big enough to hold them.

    The returned object cm = areal_cmag(obj) is a special cortical magnification object that can be
    called as cm(x, y) or cm(coord) where x and y may be lists or values and coord may be a 2D
    vector or an (n x 2) matrix. The yielded value(s) will always be in units mm^2 / deg^2.

    The following options may be provided (also in argument order):
      * retinotopy (default: 'any') specifies the retinotopy data to be used; this is passed to
        the retinotopy_data function with the mesh.
      * mask (default: None) specifies the mask that should be used; this is interpreted by a call
        to mesh.mask(mask)
      * surface_area (default: 'midgray_surface_area') specifies the vertex area property that
        should be used; this is interpreted by a call to mesh.property(surface_area) so may be
        either a property name or a list of values
      * weight (default: None) specifies additional weights that should be used; this is generally
        discouraged
      * nnearest (default: None) specifies the number of nearest neighbors to search for when
        calculating the magnification; if None, then uses int(numpy.ceil(numpy.sqrt(n) * np.log(n)))
        where n is the number of vertices included in the mask
    '''
    # First, find the retino data
    retino = retinotopy_data(mesh, retinotopy)
    # Convert from polar angle/eccen to longitude/latitude
    (ang,ecc) = as_retinotopy(retino, 'visual')
    # get the surface area
    if surface_area is None: surface_area = 'midgray_surface_area'
    if pimms.is_str(surface_area):
        if surface_area in retino: surface_area = retino[surface_area]
        elif geo.is_vset(mesh):    surface_area = mesh.property(surface_area)
        else:                      surface_area = mesh[surface_area]
    # get the weight
    if weight is not None:
        if pimms.is_str(weight):
            if weight in retino:    weight = retino[weight]
            elif geo.is_vset(mesh): weight = mesh.property(weight)
            else:                   weight = mesh[weight]
    wgt = next((retino[q] for q in ('variance_explained', 'weight') if q in retino), None)
    # Get the indices we care about
    ii = geo.to_mask(mesh, mask, indices=True)
    ii = ii[np.isfinite(ang[ii]) & np.isfinite(ecc[ii])]
    # get our nnearest
    if nnearest is None: nnearest = int(np.ceil(np.sqrt(len(ii)) * np.log(len(ii))))
    return ArealCorticalMagnification(ang[ii], ecc[ii], surface_area[ii],
                                      weight=(None if wgt is None else wgt[ii]),
                                      nnearest=nnearest)

def isoline_vmag(hemi, isolines=None, surface='midgray', min_length=2, **kw):
    '''
    isoline_vmag(hemi) calculates the visual magnification function f using the default set of
      iso-lines (as returned by neuropythy.vision.visual_isolines()). The hemi argument may
      alternately be a mesh object.
    isoline_vmag(hemi, isolns) uses the given iso-lines rather than the default ones.
    
    The return value of this funciton is a dictionary whose keys are 'tangential', 'radial', and
    'areal', and whose values are the estimated visual magnification functions. These functions
    are of the form f(x,y) where x and y can be numbers or arrays in the visual field.
    '''
    from neuropythy.util import (curry, zinv)
    from neuropythy.mri import is_cortex
    from neuropythy.vision import visual_isolines
    from neuropythy.geometry import to_mesh
    # if there's no isolines, get them
    if isolines is None: isolines = visual_isolines(hemi, **kw)
    # see if the isolines is a lazy map of visual areas; if so return a lazy map recursing...
    if pimms.is_vector(isolines.keys(), 'int'):
        f = lambda k: isoline_vmag(isolines[k], surface=surface, min_length=min_length)
        return pimms.lazy_map({k:curry(f, k) for k in six.iterkeys(isolines)})
    mesh = to_mesh((hemi, surface))
    # filter by min length
    if min_length is not None:
        isolines = {k: {kk: {kkk: [vvv[ii] for ii in iis] for (kkk,vvv) in six.iteritems(vv)}
                        for (kk,vv) in six.iteritems(v)
                        for iis in [[ii for (ii,u) in enumerate(vv['polar_angles'])
                                     if len(u) >= min_length]]
                        if len(iis) > 0}
                    for (k,v) in six.iteritems(isolines)}
    (rlns,tlns) = [isolines[k] for k in ['eccentricity', 'polar_angle']]
    if len(rlns) < 2: raise ValueError('fewer than 2 iso-eccentricity lines found')
    if len(tlns) < 2: raise ValueError('fewer than 2 iso-angle lines found')
    # grab the visual/surface lines
    ((rvlns,tvlns),(rslns,tslns)) = [[[u for lns in six.itervalues(xlns) for u in lns[k]]
                                      for xlns in (rlns,tlns)]
                                     for k in ('visual_coordinates','surface_coordinates')]
    # calculate some distances
    (rslen,tslen) = [[np.sqrt(np.sum((sx[:,:-1] - sx[:,1:])**2, 0)) for sx in slns]
                     for slns in (rslns,tslns)]
    (rvlen,tvlen) = [[np.sqrt(np.sum((vx[:,:-1] - vx[:,1:])**2, 0)) for vx in vlns]
                     for vlns in (rvlns,tvlns)]
    (rvxy, tvxy)  = [[0.5*(vx[:,:-1] + vx[:,1:]) for vx in vlns] for vlns in (rvlns,tvlns)]
    (rvlen,tvlen,rslen,tslen) = [np.concatenate(u) for u in (rvlen,tvlen,rslen,tslen)]
    (rvxy,tvxy)   = [np.hstack(vxy) for vxy in (rvxy,tvxy)]
    (rvmag,tvmag) = [vlen * zinv(slen) for (vlen,slen) in zip([rvlen,tvlen],[rslen,tslen])]
    return {k: {'visual_coordinates':vxy, 'visual_magnification': vmag,
                'visual_lengths': vlen, 'surface_lengths': slen}
            for (k,vxy,vmag,vlen,slen) in zip(['radial','tangential'], [rvxy,tvxy],
                                              [rvmag,tvmag], [rvlen,tvlen], [rslen,tslen])}

# Three methods to calculate cortical magnification:
# (1) local projection of the triangle neighborhood then comparison of path across it in the visual
#     field versus on the cortical surface
# (2) examine cortical distance along a visual field path on the cortical surface
# (3) carve out Voronoi polygons in visual space and on the cortical surface; compare areas (can
#     also do this with individual mesh triangles)

def _cmag_coord_idcs(coordinates):
    return [i for (i,(x,y)) in enumerate(zip(*coordinates))
            if (np.issubdtype(type(x), np.float) or np.issubdtype(type(x), np.int))
            if (np.issubdtype(type(y), np.float) or np.issubdtype(type(y), np.int))
            if not np.isnan(x) and not np.isnan(y)]
def _cmag_fill_result(mesh, idcs, vals):
    idcs = {idx:i for (i,idx) in enumerate(idcs)}
    return [vals[idcx[i]] if i in idcs else None for i in mesh.vertex_count]

def cmag(mesh, retinotopy='any', surface=None, to='vertices'):
    '''
    cmag(mesh) yields the neighborhood-based cortical magnification for the given mesh.
    cmag(mesh, retinotopy) uses the given retinotopy argument; this must be interpretable by
      the as_retinotopy function, or should be the name of a source (such as 'empirical' or
      'any').

    The neighborhood-based cortical magnification data is yielded as a map whose keys are 'radial',
    'tangential', 'areal', and 'field_sign'; the units of 'radial' and 'tangential' magnifications 
    are cortical-distance/degree and the units on the 'areal' magnification is
    (cortical-distance/degree)^2; the field sign has no unit.

    Note that if the retinotopy source is not given, this function will by default search for any
    source using the retinotopy_data function.

    The option surface (default None) can be provided to indicate that while the retinotopy and
    results should be formatted for the given mesh (i.e., the result should have a value for each
    vertex in mesh), the surface coordinates used to calculate areas on the cortical surface should
    come from the given surface. The surface option may be a super-mesh of mesh.

    The option to='faces' or to='vertices' (the default) specifies whether the return-values should
    be for the vertices or the faces of the given mesh. Vertex data are calculated from the face
    data by summing and averaging.
    '''
    # First, find the retino data
    if pimms.is_str(retinotopy):
        retino = retinotopy_data(mesh, retinotopy)
    else:
        retino = retinotopy
    # If this is a topology, we want to change to white surface
    if isinstance(mesh, geo.Topology): mesh = mesh.white_surface
    # Convert from polar angle/eccen to longitude/latitude
    vcoords = np.asarray(as_retinotopy(retino, 'geographical'))
    # note the surface coordinates
    if surface is None: scoords = mesh.coordinates
    else:
        scoords = surface.coordinates
        if scoords.shape[1] > mesh.vertex_count:
            scoords = scoords[:, surface.index(mesh.labels)]
    faces = mesh.tess.indexed_faces
    sx = mesh.face_coordinates
    # to understand this calculation, see this stack exchange question:
    # https://math.stackexchange.com/questions/2431913/gradient-of-angle-between-scalar-fields
    # each face has a directional magnification; we need to start with the face side lengths
    (s0,s1,s2) = np.sqrt(np.sum((np.roll(sx, -1, axis=0) - sx)**2, axis=1))
    # we want a couple other handy values:
    (s0_2,s1_2,s2_2) = (s0**2, s1**2, s2**2)
    s0_inv = zinv(s0)
    b = 0.5 * (s0_2 - s1_2 + s2_2) * s0_inv
    h = 0.5 * np.sqrt(2*s0_2*(s1_2 + s2_2) - s0_2**2 - (s1_2 - s2_2)**2) * s0_inv
    h_inv = zinv(h)
    # get the visual coordinates at each face also
    vx = np.asarray([vcoords[:,f] for f in faces])
    # we already have enough data to calculate areal magnification
    s_areas = geo.triangle_area(*sx)
    v_areas = geo.triangle_area(*vx)
    arl_mag = s_areas * zinv(v_areas)
    # calculate the gradient at each triangle; this array is dimension 2 x 2 x m where m is the
    # number of triangles; the first dimension is (vx,vy) and the second dimension is (fx,fy); fx
    # and fy are the coordinates in an arbitrary coordinate system built for each face.
    # So, to reiterate, grad is ((d(vx0)/d(fx0), d(vx0)/d(fx1)) (d(vx1)/d(fx0), d(vx1)/d(fx1)))
    dvx0_dfx = (vx[2] - vx[1]) * s0_inv
    dvx1_dfx = (vx[0] - (vx[1] + b*dvx0_dfx)) * h_inv
    grad = np.asarray([dvx0_dfx, dvx1_dfx])
    # Okay, we want to know the field signs; this is just whether the cross product of the two grad
    # vectors (dvx0/dfx and dvx1/dfx) has a positive z
    fsgn = np.sign(grad[0,0]*grad[1,1] - grad[0,1]*grad[1,0])
    # We can calculate the angle too, which is just the arccos of the normalized dot-product
    grad_norms_2 = np.sum(grad**2, axis=1)
    grad_norms = np.sqrt(grad_norms_2)
    (dvx_norms_inv, dvy_norms_inv) = zinv(grad_norms)
    ngrad = grad * ((dvx_norms_inv, dvx_norms_inv), (dvy_norms_inv, dvy_norms_inv))
    dp = np.clip(np.sum(ngrad[0] * ngrad[1], axis=0), -1, 1)
    fang = fsgn * np.arccos(dp)
    # Great; now we can calculate the drad and dtan; we have dx and dy, so we just need to
    # calculate the jacobian of ((drad/dvx, drad/dvy), (dtan/dvx, dtan/dvy))
    vx_ctr = np.mean(vx, axis=0)
    (x0, y0) = vx_ctr
    den_inv = zinv(np.sqrt(x0**2 + y0**2))
    drad_dvx = np.asarray([x0,  y0]) * den_inv
    dtan_dvx = np.asarray([-y0, x0]) * den_inv
    # get dtan and drad
    drad_dfx = np.asarray([np.sum(drad_dvx[i]*grad[:,i], axis=0) for i in [0,1]])
    dtan_dfx = np.asarray([np.sum(dtan_dvx[i]*grad[:,i], axis=0) for i in [0,1]])
    # we can now turn these into the magnitudes plus the field sign
    rad_mag = zinv(np.sqrt(np.sum(drad_dfx**2, axis=0)))
    tan_mag = zinv(np.sqrt(np.sum(dtan_dfx**2, axis=0)))
    # this is the entire result if we are doing faces only
    if to == 'faces':
        return {'radial': rad_mag, 'tangential': tan_mag, 'areal': arl_mag, 'field_sign': fsgn}
    # okay, we need to do some averaging!
    mtx = simplex_summation_matrix(mesh.tess.indexed_faces)
    cols = np.asarray(mtx.sum(axis=1), dtype=np.float)[:,0]
    cols_inv = zinv(cols)
    # for areal magnification, we want to do summation over the s and v areas then divide
    s_areas = mtx.dot(s_areas)
    v_areas = mtx.dot(v_areas)
    arl_mag = s_areas * zinv(v_areas)
    # for the others, we just average
    (rad_mag, tan_mag, fsgn) = [cols_inv * mtx.dot(x) for x in (rad_mag, tan_mag, fsgn)]
    return {'radial': rad_mag, 'tangential': tan_mag, 'areal': arl_mag, 'field_sign': fsgn}

def neighborhood_cortical_magnification(mesh, coordinates):
    '''
    neighborhood_cortical_magnification(mesh, visual_coordinates) yields a list of neighborhood-
    based cortical magnification values for the vertices in the given mesh if their visual field
    coordinates are given by the visual_coordinates matrix (must be like [x_values, y_values]). If
    either x-value or y-value of a coordinate is either None or numpy.nan, then that cortical
    magnification value is numpy.nan.
    '''
    idcs = _cmag_coord_idcs(coordinates)
    neis = mesh.tess.indexed_neighborhoods
    coords_vis = np.asarray(coordinates if len(coordinates) == 2 else coordinates.T)
    coords_srf = mesh.coordinates
    res = np.full((mesh.vertex_count, 3), np.nan, dtype=np.float)
    res = np.array([row for row in [(np.nan,np.nan,np.nan)] for _ in range(mesh.vertex_count)],
                   dtype=np.float)
    for idx in idcs:
        nei = neis[idx]
        pts_vis = coords_vis[:,nei]
        pts_srf = coords_srf[:,nei]
        x0_vis = coords_vis[:,idx]
        x0_srf = coords_srf[:,idx]
        if any(u is None or np.isnan(u) for pt in pts_vis for u in pt): continue
        # find tangential, radial, and areal magnifications
        x0col_vis = np.asarray([x0_vis]).T
        x0col_srf = np.asarray([x0_srf]).T
        # areal is easy
        voronoi_vis = (pts_vis - x0col_vis) * 0.5 + x0col_vis
        voronoi_srf = (pts_srf - x0col_srf) * 0.5 + x0col_srf
        area_vis = np.sum([geo.triangle_area(x0_vis, a, b)
                           for (a,b) in zip(voronoi_vis.T, np.roll(voronoi_vis, 1, axis=1).T)])
        area_srf = np.sum([geo.triangle_area(x0_srf, a, b)
                           for (a,b) in zip(voronoi_srf.T, np.roll(voronoi_srf, 1, axis=1).T)])
        res[idx,2] = np.inf if np.isclose(area_vis, 0) else area_srf/area_vis
        # radial and tangentual we do together because they are very similar:
        # find the intersection lines then add up their distances along the cortex
        pts_vis = voronoi_vis
        pts_srf = voronoi_srf
        segs_srf = (pts_srf, np.roll(pts_srf, -1, axis=1))
        segs_vis = (pts_vis, np.roll(pts_vis, -1, axis=1))
        segs_vis_t = np.transpose(segs_vis, (2,0,1))
        segs_srf_t = np.transpose(segs_srf, (2,0,1))
        x0norm_vis = npla.norm(x0_vis)
        if not np.isclose(x0norm_vis, 0):
            dirvecs = x0_vis / x0norm_vis
            dirvecs = np.asarray([dirvecs, [-dirvecs[1], dirvecs[0]]])
            for dirno in [0,1]:
                dirvec = dirvecs[dirno]
                line = (x0_vis, x0_vis + dirvec)
                try:
                    isects_vis = np.asarray(geo.line_segment_intersection_2D(line, segs_vis))
                    # okay, these will all be nan but two of them; they are the points we care about
                    isect_idcs = np.unique(np.where(np.logical_not(np.isnan(isects_vis)))[1])
                except Exception:
                    isect_idcs = []
                if len(isect_idcs) != 2:
                    res[idx,dirno] = np.nan
                    continue
                isects_vis = isects_vis[:,isect_idcs].T
                # we need the distance in visual space
                len_vis = npla.norm(isects_vis[0] - isects_vis[1])
                if np.isclose(len_vis, 0): res[idx,dirno] = np.inf
                else:
                    # we also need the distances on the surface: find the points by projection
                    fsegs_srf = segs_srf_t[isect_idcs]
                    fsegs_vis = segs_vis_t[isect_idcs]
                    s02lens_vis = npla.norm(fsegs_vis[:,0] - fsegs_vis[:,1], axis=1)
                    s01lens_vis = npla.norm(fsegs_vis[:,0] - isects_vis, axis=1)
                    vecs_srf = fsegs_srf[:,1] - fsegs_srf[:,0]
                    s02lens_srf = npla.norm(vecs_srf, axis=1)
                    isects_srf = np.transpose([(s01lens_vis/s02lens_vis)]) * vecs_srf \
                                 + fsegs_srf[:,0]
                    len_srf = np.sum(npla.norm(isects_srf - x0_srf, axis=1))
                    res[idx,dirno] = len_srf / len_vis
    # That's it!
    return res



####################################################################################################
# neuropythy/java/__init__.py
# The code that manages the neuropythy link to the JVM.
# By Noah C. Benson

import numpy   as np
import scipy   as sp
import numbers as num
import os, sys, gzip

from array  import array
from ..util import library_path

# Java start:
_java_port = None
_java = None

def _init_registration():
    from py4j.java_gateway import (launch_gateway, JavaGateway, GatewayParameters)
    global _java, _java_port
    if _java is not None: return
    _java_port = launch_gateway(
        classpath=os.path.join(library_path(), 'nben', 'target', 'nben-standalone.jar'),
        javaopts=['-Xmx4g'],
        die_on_exit=True)
    _java = JavaGateway(gateway_parameters=GatewayParameters(port=_java_port))

def java_link():
    if _java is None: _init_registration()
    return _java

def serialize_numpy(m, t):
    '''
    serialize_numpy(m, type) converts the numpy array m into a byte stream that can be read by the
    nben.util.Py4j Java class. The function assumes that the type of the array needn't be encoded
    in the bytearray itself. The bytearray will begin with an integer, the number of dimensions,
    followed by that number of integers (the dimension sizes themselves) then the bytes of the
    array, flattened.
    The argument type gives the type of the array to be transferred and must be 'i' for integer or
    'd' for double (or any other string accepted by array.array()).
    '''
    # Start with the header: <number of dimensions> <dim1-size> <dim2-size> ...
    header = array('i', [len(m.shape)] + list(m.shape))
    # Now, we can do the array itself, just flattened
    body = array(t, m.flatten().tolist())
    # Wrap bytes if necessary...
    if sys.byteorder != 'big':
        header.byteswap()
        body.byteswap()
    # And return the result:
    return bytearray(header.tostring() + body.tostring())

def to_java_doubles(m):
    '''
    to_java_doubles(m) yields a java array object for the vector or matrix m.
    '''
    global _java
    if _java is None: _init_registration()
    m = np.asarray(m)
    dims = len(m.shape)
    if dims > 2: raise ValueError('1D and 2D arrays supported only')
    bindat = serialize_numpy(m, 'd')
    return (_java.jvm.nben.util.Numpy.double2FromBytes(bindat) if dims == 2
            else _java.jvm.nben.util.Numpy.double1FromBytes(bindat))

def to_java_ints(m):
    '''
    to_java_ints(m) yields a java array object for the vector or matrix m.
    '''
    global _java
    if _java is None: _init_registration()
    m = np.asarray(m)
    dims = len(m.shape)
    if dims > 2: raise ValueError('1D and 2D arrays supported only')
    bindat = serialize_numpy(m, 'i')
    return (_java.jvm.nben.util.Numpy.int2FromBytes(bindat) if dims == 2
            else _java.jvm.nben.util.Numpy.int1FromBytes(bindat))

def to_java_array(m):
    '''
    to_java_array(m) yields to_java_ints(m) if m is an array of integers and to_java_doubles(m) if
    m is anything else. The numpy array m is tested via numpy.issubdtype(m.dtype, numpy.int64).
    '''
    if not hasattr(m, '__iter__'): return m
    m = np.asarray(m)
    if np.issubdtype(m.dtype, np.dtype(int).type) or all(isinstance(x, num.Integral) for x in m):
        return to_java_ints(m)
    else:
        return to_java_doubles(m)



####################################################################################################
# neuropythy/test/math.py
# Tests for the neuropythy library's util.math module.
# By Noah C. Benson

import unittest, os, sys, six, warnings, logging, pimms, torch
import numpy        as np
import scipy.sparse as sps
import pyrsistent   as pyr
import neuropythy   as ny

if sys.version_info[0] == 3: from   collections import abc as colls
else:                        import collections            as colls

class TestNeuropythyMath(unittest.TestCase):
    
    def test_info_fns(self):
        from neuropythy.math import (to_torchdtype, torchdtype_to_numpydtype, isarray, istensor,
                                     issparse, isdense)
        dts = {'int8': (torch.int8, np.int8),
               'int16': (torch.int16, np.int16),
               'int32': (torch.int32, np.int32),
               'int64': (torch.int64, np.int64),
               'float64': (torch.float64, np.float64),
               'complex128': (torch.complex128, np.complex128),
               'bool': (torch.bool, np.bool)}
        for (k, (vt,vn)) in six.iteritems(dts):
            self.assertEqual(to_torchdtype(k), vt)
            self.assertEqual(to_torchdtype(vt), vt)
            self.assertEqual(to_torchdtype(vn), vt)
            self.assertEqual(torchdtype_to_numpydtype(vt), vn)
        
        x = [1.0, 4.4, 5.1]
        y = [6,5,4,3]
        xn = np.array(x)
        yn = np.array(y)
        xt = torch.tensor(x)
        yt = torch.tensor(y)
        xs = sps.coo_matrix((x, ([0,1,2], [3,2,1])), (4,4))
        ys = sps.coo_matrix((y, ([0,1,2,3], [2,3,2,1])), (4,4))
        xp = torch.sparse_coo_tensor(([0,1,2], [3,2,1]), x, (4,4))
        yp = torch.sparse_coo_tensor(([0,1,2,3], [2,3,2,1]), y, (4,4))

        self.assertFalse(isarray(x))
        self.assertFalse(isarray(y))
        self.assertTrue(isarray(xn))
        self.assertTrue(isarray(yn))
        self.assertFalse(isarray(xt))
        self.assertFalse(isarray(yt))
        self.assertTrue(isarray(xs))
        self.assertTrue(isarray(ys))
        self.assertFalse(isarray(xp))
        self.assertFalse(isarray(yp))

        self.assertFalse(istensor(x))
        self.assertFalse(istensor(y))
        self.assertFalse(istensor(xn))
        self.assertFalse(istensor(yn))
        self.assertTrue(istensor(xt))
        self.assertTrue(istensor(yt))
        self.assertFalse(istensor(xs))
        self.assertFalse(istensor(ys))
        self.assertTrue(istensor(xp))
        self.assertTrue(istensor(yp))

        self.assertFalse(issparse(x))
        self.assertFalse(issparse(y))
        self.assertFalse(issparse(xn))
        self.assertFalse(issparse(yn))
        self.assertFalse(issparse(xt))
        self.assertFalse(issparse(yt))
        self.assertTrue(issparse(xs))
        self.assertTrue(issparse(ys))
        self.assertTrue(issparse(xp))
        self.assertTrue(issparse(yp))

        self.assertTrue(isdense(x))
        self.assertTrue(isdense(y))
        self.assertTrue(isdense(xn))
        self.assertTrue(isdense(yn))
        self.assertTrue(isdense(xt))
        self.assertTrue(isdense(yt))
        self.assertFalse(isdense(xs))
        self.assertFalse(isdense(ys))
        self.assertFalse(isdense(xp))
        self.assertFalse(isdense(yp))

    def test_create_fns(self):
        from neuropythy.math import (clone, eq, all, astensor, totensor, isarray, istensor,
                                     asarray, toarray)

        x = [1.0, 4.4, 5.1]
        y = [6,5,4,3]
        xn = np.array(x)
        yn = np.array(y)
        xt = torch.tensor(x)
        yt = torch.tensor(y)
        xs = sps.coo_matrix((x, ([0,1,2], [3,2,1])), (4,4))
        ys = sps.coo_matrix((y, ([0,1,2,3], [2,3,2,1])), (4,4))
        xp = torch.sparse_coo_tensor(([0,1,2], [3,2,1]), x, (4,4))
        yp = torch.sparse_coo_tensor(([0,1,2,3], [2,3,2,1]), y, (4,4))

        for xx in [x, xn, xt, xs, xp]:
            u = clone(xx)
            self.assertFalse(u is xx)
            self.assertTrue(all(eq(u, xx)))
        for xx in [xp, yp, xt, yt]:
            u = astensor(xx)
            self.assertTrue(u is xx)
            self.assertTrue(torch.is_tensor(u))
            u = astensor(u, dtype='float32')
            self.assertTrue(u.dtype == torch.float32)
            self.assertTrue(torch.is_tensor(u))
            self.assertTrue(all(eq(u, xx)))
            u = totensor(xx)
            self.assertFalse(u is xx)
            self.assertTrue(torch.is_tensor(u))
            self.assertFalse(isarray(u))
            self.assertTrue(istensor(u))
            self.assertTrue(all(eq(u, xx)))
            u = asarray(xx)
            self.assertFalse(u is xx)
            self.assertTrue(isinstance(u, np.ndarray) or sps.issparse(u))
            self.assertTrue(isarray(u))
            self.assertFalse(istensor(u))
            self.assertTrue(all(eq(u, xx)))
            u = toarray(xx)
            self.assertFalse(u is xx)
            self.assertTrue(isinstance(u, np.ndarray) or sps.issparse(u))
            self.assertTrue(isarray(u))
            self.assertFalse(istensor(u))
            self.assertTrue(all(eq(u, xx)))
        



####################################################################################################
# neuropythy/test/__init__.py
# Tests for the neuropythy library.
# By Noah C. Benson

import unittest, os, sys, six, warnings, logging, pimms
import numpy      as np
import pyrsistent as pyr
import neuropythy as ny

if sys.version_info[0] == 3: from   collections import abc as colls
else:                        import collections            as colls

logging.getLogger().setLevel(logging.INFO)

from .math import TestNeuropythyMath

class TestNeuropythy(unittest.TestCase):
    '''
    The TestNeuropythy class defines all the tests for the neuropythy library.
    '''

    def test_optimize(self):
        '''
        test_optimize tests the neuropythy.optimize package using the data in
          neuropythy.test.optimize.
        '''
        from neuropythy.geometry import triangle_area
        import neuropythy.optimize as opt
        import warnings
        from . import optimize as opttest
        mesh = opttest.mesh
        logging.info('neuropythy: Testing optimization package...')
        # check that this works in the first place...
        def fareas(x, f):
            x = np.asarray(x)
            (a,b,c) = [np.transpose(x[ii]) for ii in np.transpose(f)]
            return triangle_area(a, b, c)
        dif = fareas(mesh['coords'], mesh['faces'])
        sim = np.isclose(dif, opttest.mesh_face_areas, rtol=0.001)
        self.assertTrue(sim.all())
        # minimize the tiny mesh to have all triangle areas equal to 1
        m = opttest.mesh
        f = opt.sum((1.0 - opt.signed_face_areas(m['faces']))**2)
        x = f.argmin(m['coords'])
        # see if these are close to 1!
        qqq = fareas(x, m['faces']) - 1
        sim = np.isclose(qqq, 0, rtol=0, atol=0.0001)
        self.assertTrue(sim.all())

    def test_mesh(self):
        '''
        test_mesh() ensures that many general mesh properties and methods are working.
        '''
        import neuropythy.geometry as geo
        logging.info('neuropythy: Testing meshes and properties...')
        # get a random subject's mesh
        sub  = ny.data['benson_winawer_2018'].subjects['S1204']
        hem  = sub.hemis[('lh','rh')[np.random.randint(2)]]
        msh  = hem.white_surface
        # few simple things
        self.assertEqual(msh.coordinates.shape[0], 3)
        self.assertEqual(msh.tess.faces.shape[0], 3)
        self.assertEqual(msh.tess.edges.shape[0], 2)
        self.assertEqual(msh.vertex_count, msh.coordinates.shape[1])
        # face areas and edge lengths should all be non-negative
        self.assertGreaterEqual(np.min(msh.face_areas), 0)
        self.assertGreaterEqual(np.min(msh.edge_lengths), 0)
        # test the properties
        self.assertTrue('blerg' in msh.with_prop(blerg=msh.prop('curvature')).properties)
        self.assertFalse('curvature' in msh.wout_prop('curvature').properties)
        self.assertEqual(msh.properties.row_count, msh.vertex_count)
        self.assertLessEqual(np.abs(np.mean(msh.prop('curvature'))), 0.1)
        # use the property interface to grab a fancy masked property
        v123_areas = msh.property('midgray_surface_area',
                                  mask=('inf-prf_visual_area', (1,2,3)),
                                  null=0)
        v123_area = np.sum(v123_areas)
        self.assertLessEqual(v123_area, 15000)
        self.assertGreaterEqual(v123_area, 500)
        (v1_ecc, v1_rad) = msh.property(['prf_eccentricity','prf_radius'],
                                        mask=('inf-prf_visual_area', 1),
                                        weights='prf_variance_explained',
                                        weight_min=0.1,
                                        clipped=0,
                                        null=np.nan)
        wh = np.isfinite(v1_ecc) & np.isfinite(v1_rad)
        self.assertGreater(np.corrcoef(v1_ecc[wh], v1_rad[wh])[0,0], 0.5)

    def test_cmag(self):
        '''
        test_cmag() ensures that the neuropythy.vision cortical magnification function is working.
        '''
        import neuropythy.vision as vis
        logging.info('neuropythy: Testing areal cortical magnification...')
        dset = ny.data['benson_winawer_2018']
        sub = dset.subjects['S1202']
        hem = [sub.lh, sub.rh][np.random.randint(2)]
        cm = vis.areal_cmag(hem.midgray_surface, 'prf_',
                            mask=('inf-prf_visual_area', 1),
                            weight='prf_variance_explained')
        # cmag should get smaller in general
        ths = np.arange(0, 2*np.pi, np.pi/3)
        es = [0.5, 1, 2, 4]
        x = np.diff([np.mean(cm(e*np.cos(ths), e*np.sin(ths))) for e in es])
        self.assertTrue((x < 0).all())
    
    def test_interpolation(self):
        '''
        test_interpolation() performs a variety of high-level tests involving interpolation using
          neuropythy that should catch major errors to important components.
        '''
        logging.info('neuropythy: Testing interpolation...')
        def choose(coll, k): return np.random.choice(coll, k, False)
        # to do these tests, we use the builtin dataset from Benson and Winawer (2018); see also
        # help(ny.data['benson_winawer_2018']) for more information on this dataset.
        dset = ny.data['benson_winawer_2018']
        self.assertTrue(os.path.isdir(dset.cache_directory))
        # pick 1 of the subjects at random
        allsubs = [dset.subjects['S12%02d' % (s+1)] for s in range(8)]
        subs = choose(allsubs, 1)
        fsa = ny.freesurfer_subject('fsaverage')
        def check_dtypes(a,b):
            for tt in [np.integer, np.floating, np.bool_, np.complexfloating]:
                self.assertEqual(np.issubdtype(a.dtype, tt), np.issubdtype(b.dtype, tt))
        def calc_interp(hem, interhem, ps):
            for p in ps: self.assertEqual(np.sum(~np.isfinite(hem.prop(p))), 0)
            us = hem.interpolate(interhem, ps)
            for u in us: self.assertEqual(np.sum(~np.isfinite(u)), 0)
            vs = interhem.interpolate(hem, us)
            for v in vs: self.assertEqual(np.sum(~np.isfinite(v)), 0)
            return vs
        def check_interp(hem, ps, vs):
            for (p,v) in zip(ps,vs):
                logging.info('neuropythy:         * %s', p)
                p = hem.prop(p)
                self.assertEqual(len(p), len(v))
                self.assertLessEqual(np.min(p), np.min(v))
                self.assertGreaterEqual(np.max(p), np.max(v))
                check_dtypes(p, v)
                self.assertGreater(np.corrcoef(p, v)[0,0], 0.6)
        for sub in subs:
            logging.info('neuropythy: - Testing subject %s', sub.name)
            # left hemisphere should have a negative mean x-value, right a positive mean x-value
            self.assertTrue(np.mean(sub.lh.white_surface.coordinates, axis=1)[0] < 0)
            self.assertTrue(np.mean(sub.rh.pial_surface.coordinates, axis=1)[0] > 0)
            # some simple ideas: if we interpolate the properties from one subject to another and
            # then interpolate back, we should get approximately, if not exactly, the same thing
            # for this pick a couple random properties:
            ps = ['prf_variance_explained', 'inf-prf10_visual_area']
            intersub = choose(allsubs, 1)[0]
            logging.info('neuropythy:   - Testing properties %s via subject %s', ps, intersub.name)
            logging.info('neuropythy:     - Testing LH interpolation')
            vs = calc_interp(sub.lh, intersub.lh, ps)
            check_interp(sub.lh, ps, vs)
            logging.info('neuropythy:     - Testing RH interpolation')
            vs = calc_interp(sub.rh, intersub.rh, ps)
            check_interp(sub.rh, ps, vs)

    def test_path(self):
        '''
        test_path() ensures that the neuropythy.geometry.path and .path_trace data structures are
          working correctly.
        '''
        logging.info('neuropythy: Testing Path and PathTrace')
        # simple box: should have an area of ~1600 in a flatmap and something close in a sphere
        pts = [(-20,-20), (20,-20), (20,20), (-20,20)]
        # use a simple map projection
        mpj = ny.map_projection('occipital_pole', 'lh', radius=np.pi/3)
        ctx = ny.freesurfer_subject('fsaverage').lh
        trc = ny.geometry.path_trace(mpj, pts, closed=True)
        fmp = mpj(ctx)
        pth = trc.to_path(fmp)
        self.assertTrue(np.isclose(1600, pth.surface_area))
        
if __name__ == '__main__':
    unittest.main()



####################################################################################################
# neuropythy/graphics/__init__.py
# Simple tools for making matplotlib/pyplot graphics with neuropythy.
# By Noah C. Benson

'''
The neuropythy.graphics package contains definitions of the various tools for making plots with
cortical data. The primary entry point is the function cortex_plot.
'''

from .core import (
    cmap_curvature,
    cmap_polar_angle_sym, cmap_polar_angle_lh, cmap_polar_angle_rh, cmap_polar_angle,
    cmap_theta_sym, cmap_theta_lh, cmap_theta_rh, cmap_theta,
    cmap_eccentricity, cmap_log_eccentricity, cmap_radius, cmap_log_radius,
    cmap_cmag, cmap_log_cmag, label_cmap,
    vertex_curvature_color, vertex_weight,
    vertex_angle, vertex_eccen, vertex_sigma, vertex_varea,
    vertex_angle_color, vertex_eccen_color, vertex_sigma_color, vertex_varea_color,
    angle_colors, eccen_colors, sigma_colors, radius_colors, varea_colors, to_rgba,
    color_overlap, visual_field_legend, curvature_colors, cortex_plot, cortex_plot_colors,
    ROIDrawer, trace_roi, scale_for_cmap)



####################################################################################################
# registration/core.py
# Core tools for registering the cortical surface to a particular potential function
# By Noah C. Benson

import numpy  as     np
from   numpy  import pi
from   ..java import (to_java_doubles, to_java_ints, to_java_array, java_link, serialize_numpy)
from   ..     import geometry as geo
import pimms

# These are dictionaries of all the details we have about each of the possible arguments to the
# mesh_register's field argument:
_parse_field_data_types = {
    'mesh': ['newStandardMeshPotential', ['edge_scale', 1.0], ['angle_scale', 1.0], 'F', 'X'],
    'edge': {
       'harmonic':      ['newHarmonicEdgePotential',      ['scale', 1.0], ['order', 2.0], 'F', 'X'],
       'harmonic-log':  ['newHarmonicLogEdgePotential',   ['scale', 1.0], ['order', 2.0], 'F', 'X'],
       'lennard-jones': ['newLJEdgePotential',            ['scale', 1.0], ['order', 2.0], 'F', 'X'],
       'infinite-well': ['newWellEdgePotential',          ['scale', 1.0], ['order', 0.5],
                                                          ['min',   0.5], ['max',   3.0],
                                                          'E', 'X']},
    'angle': {
       'harmonic':      ['newHarmonicAnglePotential',     ['scale', 1.0], ['order', 2.0], 'F', 'X'],
       'harmonic-log':  ['newHarmonicLogAnglePotential',  ['scale', 1.0], ['order', 2.0], 'F', 'X'],
       'lennard-jones': ['newLJAnglePotential',           ['scale', 1.0], ['order', 2.0], 'F', 'X'],
       'infinite-well': ['newWellAnglePotential',         ['scale', 1.0], ['order', 0.5], 
                                                          ['min',   0.0], ['max',   pi],
                                                          'F', 'X']},
    'anchor': {
       'harmonic':      ['newHarmonicAnchorPotential',    ['scale', 1.0], ['shape', 2.0], 0,1, 'X'],
       'gaussian':      ['newGaussianAnchorPotential',    ['scale', 1.0], ['sigma', 2.0], 
                                                          ['shape', 2.0], 0, 1, 'X']},
    'mesh-field': {
       'harmonic':      ['newHarmonicMeshPotential',      ['scale', 1.0], ['order', 2.0], 0,1,2,3,4,
                                                          'X'],
       'harmonic-log':  ['newHarmonicLogMeshPotential',   ['scale', 1.0], ['order', 2.0], 0,1,2,3,4,
                                                          'X'],
       'gaussian':      ['newGaussianMeshPotential',      ['scale', 1.0], ['sigma', 0.5],
                                                          ['order', 2.0], 0,1,2,3,4, 'X']},
    'perimeter': {
       'harmonic':      ['newHarmonicPerimeterPotential', ['scale', 1.0], ['shape', 2.0],
                                                          'F', 'X']}};

def _parse_field_function_argument(argdat, args, faces, edges, coords):
    # first, see if this is an easy one...
    if argdat == 'F':
        return faces
    elif argdat == 'X':
        return coords
    elif argdat == 'E':
        return edges
    elif pimms.is_int(argdat):
        return to_java_array(args[argdat])
    # okay, none of those; must be a list with a default arg
    argname = argdat[0]
    argdflt = argdat[1]
    # see if we can find such an arg...
    for i in range(len(args)):
        if pimms.is_str(args[i]) and args[i].lower() == argname.lower():
            return (args[i+1] if pimms.is_number(args[i+1]) else to_java_array(args[i+1]))
    # did not find the arg; use the default:
    return argdflt

def _parse_field_argument(instruct, faces, edges, coords):
    _java = java_link()
    if pimms.is_str(instruct):
        insttype = instruct
        instargs = []
    elif hasattr(instruct, '__iter__'):
        insttype = instruct[0]
        instargs = instruct[1:]
    else:
        raise RuntimeError('potential field instruction must be list/tuple-like or a string')
    # look this type up in the types data:
    insttype = insttype.lower()
    if insttype not in _parse_field_data_types:
        raise RuntimeError('Unrecognized field data type: ' + insttype)
    instdata = _parse_field_data_types[insttype]
    # if the data is a dictionary, we must parse on the next arg
    if pimms.is_map(instdata):
        shape_name = instargs[0].lower()
        instargs = instargs[1:]
        if shape_name not in instdata:
            raise RuntimeError('Shape ' + shape_name + ' not supported for type ' + insttype)
        instdata = instdata[shape_name]
    # okay, we have a list of instructions... find the java method we are going to call...
    java_method = getattr(_java.jvm.nben.mesh.registration.Fields, instdata[0])
    # and parse the arguments into a list...
    java_args = [_parse_field_function_argument(a, instargs, faces, edges, coords)
                 for a in instdata[1:]]
    # and call the function...
    return java_method(*java_args)

# parse a field potential argument and return a java object that represents it
def _parse_field_arguments(arg, faces, edges, coords):
    '''See mesh_register.'''
    if not hasattr(arg, '__iter__'):
        raise RuntimeError('field argument must be a list-like collection of instructions')
    pot = [_parse_field_argument(instruct, faces, edges, coords) for instruct in arg]
    # make a new Potential sum unless the length is 1
    if len(pot) <= 1:
        return pot[0]
    else:
        sp = java_link().jvm.nben.mesh.registration.Fields.newSum()
        for field in pot: sp.addField(field)
        return sp

def java_potential_term(mesh, instructions):
    '''
    java_potential_term(mesh, instructions) yields a Java object that implements the potential field
      described in the given list of instructions. Generally, this should not be invoked directly
      and should only be called by mesh_register. Note: this expects a single term's description,
      not a series of descriptions.
    '''
    faces  = to_java_ints(mesh.indexed_faces)
    edges  = to_java_ints(mesh.indexed_edges)
    coords = to_java_doubles(mesh.coordinates)
    return _parse_field_arguments([instructions], faces, edges, coords)
    
# The mesh_register function
def mesh_register(mesh, field, max_steps=2000, max_step_size=0.05, max_pe_change=1,
                  method='random', return_report=False, initial_coordinates=None):
    '''
    mesh_register(mesh, field) yields the mesh that results from registering the given mesh by
    minimizing the given potential field description over the position of the vertices in the
    mesh. The mesh argument must be a Mesh object (see neuropythy.geometry) such as can be read
    from FreeSurfer using the neuropythy.freesurfer_subject function. The field argument must be
    a list of field names and arguments; with the exception of 'mesh' (or 'standard'), the 
    arguments must be a list, the first element of which is the field type name, the second
    element of which is the field shape name, and the final element of which is a dictionary of
    arguments accepted by the field shape.

    The following are valid field type names:
      * 'mesh' : the standard mesh potential, which includes an edge potential, an angle
        potential, and a perimeter potential. Accepts no arguments, and must be passed as a
        single string instead of a list.
      * 'edge': an edge potential field in which the potential is a function of the change in the
        edge length, summed over each edge in the mesh.
      * 'angle': an angle potential field in which the potential is a function of the change in
        the angle measure, summed over all angles in the mesh.
      * 'perimeter': a potential that depends on the vertices on the perimeter of a 2D mesh
        remaining in place; the potential changes as a function of the distance of each perimeter
        vertex from its reference position.
      * 'anchor': a potential that depends on the distance of a set of vertices from fixed points
        in space. After the shape name second argument, an anchor must be followed by a list of
        vertex ids then a list of fixed points to which the vertex ids are anchored:
        ['anchor', shape_name, vertex_ids, fixed_points, args...].

    The following are valid shape names:
      * 'harmonic': a harmonic function with the form (c/q) * abs(x - x0)^q.
        Parameters: 
          * 'scale', the scale parameter c; default: 1.
          * 'order', the order parameter q; default: 2.
      * 'Lennard-Jones': a Lennard-Jones function with the form c (1 + (r0/r)^q - 2(r0/r)^(q/2));
        Parameters:
          * 'scale': the scale parameter c; default: 1. 
          * 'order': the order parameter q; default: 2.
      * 'Gaussian': A Gaussian function with the form c (1 - exp(-0.5 abs((x - x0)/s)^q))
        Parameters:
          * 'scale': the scale parameter c; default: 1.
          * 'order': the order parameter q; default: 2.
          * 'sigma': the standard deviation parameter s; default: 1.
      * 'infinite-well': an infinite well function with the form 
        c ( (((x0 - m)/(x - m))^q - 1)^2 + (((M - x0)/(M - x))^q - 1)^2 )
        Parameters:
          * 'scale': the scale parameter c; default: 1.
          * 'order': the order parameter q; default: 0.5.
          * 'min': the minimum value m; default: 0.
          * 'max': the maximum value M; default: pi.

    Options: The following optional arguments are accepted.
      * max_steps (default: 2000) the maximum number of steps to minimize for.
      * max_step_size (default: 0.1) the maximum distance to allow a vertex to move in a single
        minimization step.
      * max_pe_change: the maximum fraction of the initial potential value that the minimizer
        should minimize away before returning; i.e., 0 indicates that no minimization should be
        allowed while 0.9 would indicate that the minimizer should minimize until the potential
        is 10% or less of the initial potential.
      * return_report (default: False) indicates that instead of returning the registered data,
        mesh_register should instead return the Java Minimizer.Report object (for debugging).
      * method (default: 'random') specifies the search algorithm used; available options are 
        'random', 'nimble', and 'pure'. Generally all options will converge on a similar solution,
        but usually 'random' is fastest. The 'pure' option uses the nben library's step function,
        which performs straight-forward gradient descent. The 'nimble' option performs a gradient
        descent in which subsets of vertices in the mesh that have the highest gradients during the
        registration are updated more often than those vertices with small gradients; this can
        sometimes but not always increase the speed of the minimization. Note that instead of
        'nimble', one may alternately provide ('nimble', k) where k is the number of partitions that
        the vertices should be sorted into (by partition). 'nimble' by itself is equivalent to 
        ('nimble', 4). Note also that a single step of nimble minimization is equivalent to 2**k
        steps of 'pure' minimization. Finally, the 'random' option uses the nben library's
        randomStep function, which is a gradient descent algorithm that moves each vertex in the
        direction of its negative gradient during each step but which randomizes the length of the
        gradient at each individual vertex by drawing from an exponential distribution centered at
        the vertex's actual gradient length. In effect, this can prevent vertices with very large
        gradients from dominating the minimization and often results in the best results.
      * initial_coordinates (default: None) specifies the start coordinates of the registration;
        if None, uses those in the given mesh, which is generally desired.

    Examples:
      registered_mesh = mesh_register(
         mesh,
         [['edge', 'harmonic', 'scale', 0.5], # slightly weak edge potential
          ['angle', 'infinite-well'], # default arguments for an infinite-well angle potential
          ['anchor', 'Gaussian', [1, 10, 50], [[0.0, 0.0], [1.1, 1.1], [2.2, 2.2]]]],
         max_step_size=0.05,
         max_steps=10000)
    '''
    # Sanity checking:
    # First, make sure that the arguments are all okay:
    if not isinstance(mesh, geo.Mesh):
        raise RuntimeError('mesh argument must be an instance of neuropythy.geometry.Mesh')
    if not pimms.is_vector(max_steps): max_steps = [max_steps]
    for ms in max_steps:
        if not pimms.is_int(ms) or ms < 0:
            raise RuntimeError('max_steps argument must be a positive integer')
    if not pimms.is_vector(max_step_size): max_step_size = [max_step_size]
    for mss in max_step_size:
        if not pimms.is_number(mss) or mss <= 0:
            raise RuntimeError('max_step_size must be a positive number')
    if not pimms.is_number(max_pe_change) or max_pe_change <= 0 or max_pe_change > 1:
        raise RuntimeError('max_pe_change must be a number x such that 0 < x <= 1')
    if pimms.is_vector(method):
        if method[0].lower() == 'nimble' and len(method) > 1 and not pimms.is_str(method[1]):
            method = [method]
    else: method = [method]
    if initial_coordinates is None:
        init_coords = mesh.coordinates
    else:
        init_coords = np.asarray(initial_coordinates)
        if init_coords.shape[0] != mesh.coordinates.shape[0]:
            init_coords = init_coords.T
    # If steps is 0, we can skip most of this...
    if np.sum(max_steps) == 0:
        if return_report: return None
        else: return init_coords
    # Otherwise, we run at least some minimization
    max_pe_change = float(max_pe_change)
    nrounds = len(max_steps)
    if nrounds > 1:
        if len(max_step_size) == 1: max_step_size = [max_step_size[0] for _ in max_steps]
        if len(method) == 1:        method        = [method[0]        for _ in max_steps]
    # Parse the field argument.
    faces  = to_java_ints(mesh.tess.indexed_faces)
    edges  = to_java_ints(mesh.tess.indexed_edges)
    coords = to_java_doubles(mesh.coordinates)
    init_coords = coords if init_coords is mesh.coordinates else to_java_doubles(init_coords)
    potential = _parse_field_arguments(field, faces, edges, coords)
    # Okay, that's basically all we need to do the minimization...
    rep = []
    for (method,max_step_size,max_steps) in zip(method, max_step_size, max_steps):
        minimizer = java_link().jvm.nben.mesh.registration.Minimizer(potential, init_coords)
        max_step_size = float(max_step_size)
        max_steps = int(max_steps)
        if pimms.is_str(method):
            method = method.lower()
            if method == 'nimble': k = 4
            else:                  k = 0
        else:
            k = method[1]
            method = method[0].lower()
        if method == 'pure':
            r = minimizer.step(max_pe_change, max_steps, max_step_size)
        elif method == 'random':
            # if k is -1, we do the inverse version where we draw from the 1/mean distribution
            r = minimizer.randomStep(max_pe_change, max_steps, max_step_size, k == -1)
        elif method == 'nimble':
            r = minimizer.nimbleStep(max_pe_change, max_steps, max_step_size, int(k))
        else:
            raise ValueError('Unrecognized method: %s' % method)
        rep.append(r)
        init_coords = minimizer.getX()
    # Return the report if requested
    if return_report:
        return rep
    else:
        result = init_coords
        return np.asarray([[x for x in row] for row in result])

# The topology and registration stuff is below:
class JavaTopology:
    '''
    JavaTopology(triangles, registrations) creates a topology object object with the given triangle
    mesh, defined by a 3xn matrix of triangle indices, and with the registration coordinate matrices
    given in the dictionary registrations. This class should only be instantiated by the neuropythy
    library and should generally not be constructed directly. See Hemisphere.topology objects to
    access a subject's topologies.
    '''
    def __init__(self, triangles, registrations):
        # First: make a java object for the topology:
        faces = serialize_numpy(triangles.T, 'i')
        topo = java_link().jvm.nben.geometry.spherical.MeshTopology.fromBytes(faces)
        # Okay, make our registration dictionary
        d = {k: topo.registerBytes(serialize_numpy(v, 'd'))
             for (k,v) in registrations.iteritems()}
        # That's all really
        self.__dict__['_java_object'] = topo
        self.__dict__['registrations'] = d
    def __getitem__(self, attribute):
        return self.registrations[attribute]
    def __setitem__(self, attribute, dat):
        self.registrations[attribute] = self._java_object.registerBytes(serialize_numpy(dat, 'd'))
    def keys(self):
        return self.registrations.keys()
    def iterkeys(self):
        return self.registrations.iterkeys()
    def values(self):
        return self.registrations.values()
    def itervalues(self):
        return self.registrations.itervalues()
    def items(self):
        return self.registrations.items()
    def iteritems(self):
        return self.registrations.iteritems()
    def __len__(self):
        return len(self.registrations)
    
    # These let us interpolate...
    def interpolate(fromtopo, data, order=2, fill=None):
        usable_keys = []
        for k in registrations.iterkeys():
            if k in fromtopo.registrations:
                usable_keys.append(k)
        if not usable_keys:
            raise RuntimeError('no registration found that links topologies')
        the_key = usable_keys[0]
        # Prep the data into java arrays
        jmask = serialize_numpy(np.asarray([1 if d is not None else 0 for d in data]), 'd')
        jdata = serialize_numpy(np.asarray([d if d is not None else 0 for d in data]), 'd')
        # okay, next step is to call out to the java...
        maskres = self._java_object.interpolateBytes(
            fromtopo.registrations[the_key],
            self.registrations[the_key].coordinates,
            order, jdata)
        datares = self._java_object.interpolateBytes(
            fromtopo.registrations[the_key],
            self.registrations[the_key].coordinates,
            order, jmask)
        # then interpret the results...
        return [datares[i] if maskres[i] == 1 else fill for i in range(len(maskres))]




####################################################################################################
# neuropythy/optimize/__init__.py
# Initialization code for neuropythy's optimization module.
# By Noah C. Benson

from .core import (fapply, finto,
                   PotentialFunction, is_potential, to_potential,
                   is_const_potential, const_potential, const, identity, is_identity_potential,
                   compose, part, exp, exp2, power, sqrt, log, log2, log10, erf, sum, dot,
                   cos, sin, tan, sec, csc, cot, asin, acos, atan, atan2,
                   piecewise, cos_well, cos_edge, abs, sign, gaussian, sigmoid,
                   row_norms, col_norms, distances,
                   signed_face_areas, face_areas)
